#
# Copyright (c), 2021-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Define protocols for type annotation of XPath related objects.
"""
from collections.abc import Callable, Hashable, Iterator, Iterable, \
    ItemsView, Mapping, Sequence, Sized
from typing import overload, Any, Optional, Protocol, Union, TypeVar
from xml.etree.ElementTree import Element, ElementTree

from collections.abc import MutableMapping
from elementpath.aliases import NamespacesType, NsmapType

_T = TypeVar("_T")
_AnyStr = Union[str, bytes]


class LxmlQNameProtocol(Protocol):
    localname: _AnyStr
    namespace: _AnyStr
    text: _AnyStr


LxmlKeyType = Union[str, bytes, LxmlQNameProtocol]


class LxmlAttribProtocol(Protocol):
    """A minimal protocol for attributes of lxml Element objects."""
    def get(self, *args: Any, **kwargs: Any) -> Optional[str]: ...

    def items(self) -> Sequence[tuple[Any, Any]]: ...

    def __contains__(self, key: Any) -> bool: ...

    def __getitem__(self, key: Any) -> Any: ...

    def __iter__(self) -> Iterator[Any]: ...

    def __len__(self) -> int: ...


class ElementProtocol(Sized, Hashable, Protocol):
    """A protocol for generic ElementTree elements."""

    def __iter__(self) -> Iterator['ElementProtocol']: ...

    def find(
            self, path: str, namespaces: Optional[dict[str, str]] = ...
    ) -> Optional['ElementProtocol']: ...
    def iter(self, tag: Optional[str] = ...) -> Iterator['ElementProtocol']: ...

    @overload
    def get(self, key: str) -> Optional[str]: ...

    @overload
    def get(self, key: str, default: _T) -> Union[str, _T]: ...

    def get(self, key: str, default: Optional[_T] = None) -> Union[str, _T, None]: ...

    @property
    def tag(self) -> Union[str, Callable[[], 'ElementProtocol']]: ...

    @property
    def text(self) -> Optional[str]: ...

    @property
    def tail(self) -> Optional[str]: ...

    @property
    def attrib(self) -> 'AttribType': ...


class EtreeElementProtocol(ElementProtocol, Protocol):
    """A protocol for xml.etree.ElementTree elements."""
    def __iter__(self) -> Iterator['EtreeElementProtocol']: ...

    def find(
            self, path: str, namespaces: Optional[dict[str, str]] = ...
    ) -> Optional['EtreeElementProtocol']: ...
    def iter(self, tag: Optional[str] = ...) -> Iterator['EtreeElementProtocol']: ...

    @property
    def attrib(self) -> dict[str, str]: ...


class LxmlElementProtocol(ElementProtocol, Protocol):
    """A protocol for lxml.etree elements."""
    def __iter__(self) -> Iterator['LxmlElementProtocol']: ...

    def find(
            self, path: str, namespaces: Optional[MutableMapping[str, str]] = ...
    ) -> Optional['LxmlElementProtocol']: ...
    def iter(self, tag: Optional[str] = ...) -> Iterator['LxmlElementProtocol']: ...

    def getroottree(self) -> 'LxmlDocumentProtocol': ...
    def getnext(self) -> Optional['LxmlElementProtocol']: ...
    def getparent(self) -> Optional['LxmlElementProtocol']: ...
    def getprevious(self) -> Optional['LxmlElementProtocol']: ...
    def itersiblings(self, tag: Optional[str] = ..., *tags: str,
                     preceding: bool = False) -> Iterable['LxmlElementProtocol']: ...

    @property
    def nsmap(self) -> NsmapType: ...

    @property
    def attrib(self) -> LxmlAttribProtocol: ...


class DocumentProtocol(Hashable, Protocol):
    def getroot(self) -> Optional[ElementProtocol]: ...
    def parse(self, source: Any, *args: Any, **kwargs: Any) -> ElementProtocol: ...
    def iter(self, tag: Optional[str] = ...) -> Iterator[ElementProtocol]: ...


class LxmlDocumentProtocol(Hashable, Protocol):
    def getroot(self) -> Optional[LxmlElementProtocol]: ...
    def parse(self, source: Any, *args: Any, **kwargs: Any) -> LxmlElementProtocol: ...
    def iter(self, tag: Optional[str] = ...) -> Iterator[LxmlElementProtocol]: ...


class XsdValidatorProtocol(Hashable, Protocol):
    def is_matching(self, name: Optional[str],
                    default_namespace: Optional[str] = None) -> bool: ...

    @property
    def name(self) -> Optional[str]: ...

    @property
    def xsd_version(self) -> str: ...

    @property
    def maps(self) -> 'GlobalMapsProtocol': ...


class XsdComponentProtocol(XsdValidatorProtocol, Protocol):

    @property
    def parent(self) -> Optional['XsdComponentProtocol']: ...

    @property
    def ref(self) -> Optional['XsdComponentProtocol']: ...


class XsdTypeProtocol(XsdComponentProtocol, Protocol):

    def is_simple(self) -> bool:
        """Returns `True` if it's a simpleType instance, `False` if it's a complexType."""
        ...

    def is_empty(self) -> bool:
        """
        Returns `True` if it's a simpleType instance or a complexType with empty content,
        `False` otherwise.
        """
        ...

    def has_simple_content(self) -> bool:
        """
        Returns `True` if it's a simpleType instance or a complexType with simple content,
        `False` otherwise.
        """
        ...

    def has_mixed_content(self) -> bool:
        """
        Returns `True` if it's a complexType with mixed content, `False` otherwise.
        """
        ...

    def is_element_only(self) -> bool:
        """
        Returns `True` if it's a complexType with element-only content, `False` otherwise.
        """
        ...

    def is_atomic(self) -> bool:
        """Returns `True` if the instance is an atomic simpleType, `False` otherwise."""
        ...

    def is_list(self) -> bool:
        """Returns `True` if the instance is a list simpleType, `False` otherwise."""
        ...

    def is_union(self) -> bool:
        """Returns `True` if the instance is a union simpleType, `False` otherwise."""
        ...

    def is_key(self) -> bool:
        """Returns `True` if it's a simpleType derived from xs:ID, `False` otherwise."""
        ...

    def is_qname(self) -> bool:
        """Returns `True` if it's a simpleType derived from xs:QName, `False` otherwise."""
        ...

    def is_notation(self) -> bool:
        """Returns `True` if it's a simpleType derived from xs:NOTATION, `False` otherwise."""
        ...

    @overload
    def is_valid(self, obj: Any, use_defaults: bool = True,
                 namespaces: Optional[NamespacesType] = None,
                 *args: Any, **kwargs: Any) -> bool: ...

    @overload
    def is_valid(self, obj: Any, *args: Any, **kwargs: Any) -> bool: ...

    def is_valid(self, obj: Any, *args: Any, **kwargs: Any) -> bool:
        """
        Validates an XML object node using the XSD type. The argument *obj* is an element
        for complex type nodes or a text value for simple type nodes. Returns `True` if
        the argument is valid, `False` otherwise.
        """
        ...

    @overload
    def validate(self, obj: Any, use_defaults: bool = True,
                 namespaces: Optional[NamespacesType] = None,
                 *args: Any, **kwargs: Any) -> None: ...

    @overload
    def validate(self, obj: Any, *args: Any, **kwargs: Any) -> None: ...

    def validate(self, obj: Any, *args: Any, **kwargs: Any) -> None:
        """
        Validates an XML object node using the XSD type. The argument *obj* is an element
        for complex type nodes or a text value for simple type nodes. Raises a `ValueError`
        compatible exception (a `ValueError` or a subclass of it) if the argument is not valid.
        """
        ...

    def decode(self, obj: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Decodes an XML object node using the XSD type. The argument *obj* is an element
        for complex type nodes or a text value for simple type nodes. Raises a `ValueError`
        or a `TypeError` compatible exception if the argument it's not valid.
        """
        ...

    @property
    def root_type(self) -> 'XsdTypeProtocol':
        """
        The type at base of the definition of the XSD type. For a special type is the type
        itself. For an atomic type is the primitive type. For a list is the primitive type
        of the item. For a union is the base union type. For a complex type is xs:anyType.
        """
        ...

    @property
    def simple_type(self) -> Optional['XsdTypeProtocol']:
        """
        The instance if it's a simpleType instance or the simpleType instance used for
        deriving a complexType with simple content, `None` otherwise.
        """
        ...

    @property
    def model_group(self) -> Optional['XsdGroupProtocol']:
        """
        A model group if it's a complexType with mixed or element-only content, `None` otherwise.
        """
        ...


class XsdComplexTypeProtocol(XsdComponentProtocol, Protocol):

    @property
    def attributes(self) -> 'XsdAttributeGroupProtocol': ...


class XsdGroupProtocol(XsdComponentProtocol, Protocol):

    def iter_elements(self) -> Iterator['XsdElementProtocol']: ...


class XsdAttributeProtocol(XsdComponentProtocol, Protocol):

    @property
    def type(self) -> XsdTypeProtocol: ...


class XsdAnyAttributeProtocol(XsdComponentProtocol, Protocol):

    @property
    def type(self) -> None: ...


XsdAttributeProtocolType = Union[XsdAnyAttributeProtocol, XsdAttributeProtocol]


class XsdAttributeGroupProtocol(XsdComponentProtocol, Protocol):

    def get(self, key: Optional[str], /) -> Optional[XsdAttributeProtocolType]: ...

    def items(self) -> ItemsView[Optional[str], XsdAttributeProtocol]: ...

    def __contains__(self, key: Optional[str]) -> bool: ...

    def __getitem__(self, key: Optional[str]) -> XsdAttributeProtocol: ...

    def __iter__(self) -> Iterator[Optional[str]]: ...

    def __len__(self) -> int: ...

    def __hash__(self) -> int: ...


class XsdElementProtocol(XsdComponentProtocol, ElementProtocol, Protocol):

    @property
    def name(self) -> Optional[str]: ...

    @property
    def type(self) -> Optional[XsdTypeProtocol]: ...

    @property
    def attrib(self) -> XsdAttributeGroupProtocol: ...

    def __iter__(self) -> Iterator['XsdElementProtocol']: ...

    def find(self, path: str, namespaces: Optional[NamespacesType] = ...) \
        -> Union['XsdElementProtocol', None]: ...


class GlobalMapsProtocol(Protocol):

    @property
    def types(self) -> Mapping[str, XsdTypeProtocol]: ...

    @property
    def attributes(self) -> Mapping[str, XsdAttributeProtocol]: ...

    @property
    def elements(self) -> Mapping[str, XsdElementProtocol]: ...

    @property
    def substitution_groups(self) -> Mapping[str, set[Any]]: ...


class XsdSchemaProtocol(XsdValidatorProtocol, Protocol):

    @property
    def validity(self) -> str: ...

    @property
    def validation_attempted(self) -> str: ...

    @property
    def tag(self) -> str: ...

    @property
    def attrib(self) -> MutableMapping[Optional[str], 'XsdAttributeProtocol']: ...

    def __iter__(self) -> Iterator['XsdXPathNodeType']: ...

    def find(self, path: str, namespaces: Optional[NamespacesType] = ...) \
        -> Union['XsdSchemaProtocol', 'XsdElementProtocol', None]: ...


XsdXPathNodeType = Union[XsdSchemaProtocol, XsdElementProtocol]
DocumentType = Union[ElementTree, DocumentProtocol]
ElementType = Union[Element, ElementProtocol]
SchemaElemType = Union[XsdSchemaProtocol, XsdElementProtocol]
CommentType = Union[Element, ElementProtocol]
ProcessingInstructionType = Union[Element, ElementProtocol]
AttribType = Union[
    MutableMapping[str, Any],
    MutableMapping[Optional[str], Any],
    LxmlAttribProtocol,
    XsdAttributeGroupProtocol
]

__all__ = ['ElementProtocol', 'EtreeElementProtocol', 'LxmlAttribProtocol',
           'LxmlElementProtocol', 'DocumentProtocol', 'LxmlDocumentProtocol',
           'XsdValidatorProtocol', 'XsdComponentProtocol', 'XsdTypeProtocol',
           'XsdAttributeProtocol', 'XsdAttributeGroupProtocol', 'XsdElementProtocol',
           'GlobalMapsProtocol', 'XsdSchemaProtocol', 'DocumentType', 'ElementType',
           'SchemaElemType', 'CommentType', 'ProcessingInstructionType',
           'AttribType', 'XsdXPathNodeType']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XSD atomic datatypes subpackage. Includes a class for UntypedAtomic data and
classes for other XSD built-in types. This subpackage raises only built-in
exceptions in order to be reusable by other packages.
"""
from .any_types import builtin_atomic_types, AtomicTypeMeta, AnyType, \
    AnySimpleType, AnyAtomicType
from .numeric import Float, Float10, Integer, Int, Long, \
    NegativeInteger, PositiveInteger, NonNegativeInteger, \
    NonPositiveInteger, Short, Byte, UnsignedByte, UnsignedInt, \
    UnsignedLong, UnsignedShort
from .untyped import UntypedAtomic
from .qname import AbstractQName, QName, Notation
from .string import NormalizedString, XsdToken, Name, NCName, \
    NMToken, Id, Idref, Language, Entity
from .lists import builtin_list_types, ListType, NMTokens, Idrefs, Entities
from .uri import AnyURI
from .binary import AbstractBinary, Base64Binary, HexBinary
from .datetime import AbstractDateTime, DateTime10, DateTime, DateTimeStamp, \
    Date10, Date, GregorianDay, GregorianMonth, GregorianYear, GregorianYear10, \
    GregorianMonthDay, GregorianYearMonth, GregorianYearMonth10, Time, Timezone, \
    Duration, DayTimeDuration, YearMonthDuration
from .proxies import ErrorProxy, BooleanProxy, DecimalProxy, DoubleProxy, \
    DoubleProxy10, StringProxy, NumericProxy, ArithmeticProxy
from ..sequences import XSequence

###
# Alias kept for backward compatibility, will be removed in v6.0.
OrderedDateTime = AbstractDateTime

__all__ = ['AbstractBinary', 'AbstractDateTime', 'ListType', 'AbstractQName',
           'AnyAtomicType', 'AnyURI', 'ArithmeticProxy', 'AtomicTypeMeta', 'Base64Binary',
           'BooleanProxy', 'AtomicTypeMeta', 'Byte', 'Date', 'Date10', 'DateTime',
           'DateTime10', 'DateTimeStamp', 'DayTimeDuration', 'DecimalProxy', 'DoubleProxy',
           'DoubleProxy10', 'Duration', 'Entities', 'Entity', 'ErrorProxy',
           'Float', 'Float10', 'GregorianDay', 'GregorianMonth', 'GregorianMonthDay',
           'GregorianYear', 'GregorianYear10', 'GregorianYearMonth', 'GregorianYearMonth10',
           'HexBinary', 'Id', 'Idref', 'Idrefs', 'Int', 'Integer', 'Language', 'Long',
           'NCName', 'NMToken', 'NMTokens', 'Name', 'NegativeInteger', 'NonNegativeInteger',
           'NonPositiveInteger', 'NormalizedString', 'Notation', 'NumericProxy',
           'OrderedDateTime', 'PositiveInteger', 'QName', 'Short', 'StringProxy', 'Time',
           'Timezone', 'UnsignedByte', 'UnsignedInt', 'UnsignedLong', 'UnsignedShort',
           'UntypedAtomic', 'XsdToken', 'YearMonthDuration', 'ListType', 'XSequence',
           'builtin_atomic_types', 'builtin_list_types', 'AnyType', 'AnySimpleType']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 1.0 implementation - part 3 (functions)
"""
import math
import decimal
from collections.abc import Iterator
from typing import Any

import elementpath.aliases as ta

from elementpath.namespaces import XML_ID, XML_LANG
from elementpath.datatypes import AnyURI, Float, DayTimeDuration, YearMonthDuration, \
    StringProxy, AnyAtomicType, Duration
from elementpath.helpers import get_double
from elementpath.xpath_nodes import XPathNode, ElementNode, TextNode, CommentNode, \
    ProcessingInstructionNode, DocumentNode, EtreeElementNode
from elementpath.xpath_context import XPathSchemaContext
from elementpath.xpath_tokens import XPathFunction

from ._xpath1_operators import XPath1Parser

__all__ = ['XPath1Parser']

method = XPath1Parser.method
function = XPath1Parser.function


###
# Kind tests (for matching of node types in XPath 1.0 or sequence types in XPath 2.0)
@method(function('node', nargs=0, label='kind test'))
def select__node_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[XPathNode]:
    if context is None:
        raise self.missing_context()

    for item in context.iter_children_or_self():
        if isinstance(item, XPathNode):
            if not isinstance(item, DocumentNode) or item is context.root:
                yield item


@method('node')
def nud__item_sequence_type(self: XPathFunction) -> XPathFunction:
    XPathFunction.nud(self)
    if self.parser.next_token.symbol in ('*', '+', '?'):
        self.occurrence = self.parser.next_token.symbol
        self.parser.advance()
    return self


@method(function('processing-instruction', nargs=(0, 1), bp=79, label='kind test'))
def select__pi_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ProcessingInstructionNode]:
    if context is None:
        raise self.missing_context()

    for item in context.iter_children_or_self():
        if isinstance(item, ProcessingInstructionNode):
            if not self:
                yield item
            elif isinstance(name := self[0].value, str):
                if item.name == ' '.join(name.strip().split()):
                    yield item


@method('processing-instruction')
def nud__pi_kind_test(self: XPathFunction) -> XPathFunction:
    self.parser.advance('(')
    if self.parser.next_token.symbol != ')':
        self.parser.next_token.expected('(name)', '(string)')
        self[0:] = self.parser.expression(5),
    self.parser.advance(')')
    return self


@method(function('comment', nargs=0, label='kind test'))
def select__comment_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[CommentNode]:
    if context is None:
        raise self.missing_context()

    for item in context.iter_children_or_self():
        if isinstance(item, CommentNode):
            yield item


@method(function('text', nargs=0, label='kind test'))
def select__text_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[TextNode]:
    if context is None:
        raise self.missing_context()

    for item in context.iter_children_or_self():
        if isinstance(item, TextNode):
            yield item


###
# Node set functions
@method(function('last', nargs=0, sequence_types=('xs:integer',)))
def evaluate__last_function(self: XPathFunction, context: ta.ContextType = None) -> int:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()
    elif context.size is None:
        raise self.missing_context("Context size is undefined")
    return context.size


@method(function('position', nargs=0,
                 sequence_types=('xs:integer',)))
def evaluate__position(self: XPathFunction, context: ta.ContextType = None) -> int:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()
    return context.position


@method(function('count', nargs=1, sequence_types=('item()*', 'xs:integer')))
def evaluate__count(self: XPathFunction, context: ta.ContextType = None) -> int:
    return len([x for x in self[0].select(self.context or context)])


@method(function('id', nargs=1, sequence_types=('xs:string*', 'element()*')))
def select__id(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ElementNode]:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    value = self[0].evaluate(context)
    item = context.item
    if item is None:
        item = context.root

    if isinstance(item, (ElementNode, DocumentNode)):
        for element in item.iter_descendants():
            if isinstance(element, EtreeElementNode) and element.value.get(XML_ID) == value:
                yield element


@method(function('name', nargs=(0, 1), sequence_types=('node()?', 'xs:string')))
@method(function('local-name', nargs=(0, 1), sequence_types=('node()?', 'xs:string')))
@method(function('namespace-uri', nargs=(0, 1), sequence_types=('node()?', 'xs:anyURI')))
def evaluate__name_related_functions(self: XPathFunction, context: ta.ContextType = None) \
        -> str | AnyURI:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    arg = self.get_argument(context, default_to_context=True)
    if arg is None:
        return ''
    elif not isinstance(arg, XPathNode):
        raise self.error('XPTY0004')

    name = arg.name
    if name is None:
        return ''

    symbol = self.symbol
    if symbol == 'name':
        node_name = arg.node_name
        if node_name is None:
            return ''
        return node_name.qname
    elif symbol == 'local-name':
        return name if not name or name[0] != '{' else name.split('}')[1]
    elif self.parser.version == '1.0':
        return '' if not name or name[0] != '{' else name.split('}')[0][1:]
    else:
        return AnyURI('') if not name or name[0] != '{' else AnyURI(name.split('}')[0][1:])


###
# String functions
@method(function('string', nargs=(0, 1), sequence_types=('item()?', 'xs:string')))
def evaluate__string(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    if not self:
        if context is None:
            raise self.missing_context()
        return self.string_value(context.item)
    return self.string_value(self.get_argument(context))


@method(function('contains', nargs=2,
                 sequence_types=('xs:string?', 'xs:string?', 'xs:boolean')))
def evaluate__contains(self: XPathFunction, context: ta.ContextType = None) -> bool:
    if self.context is not None:
        context = self.context

    arg1 = self.get_argument(context, default='', cls=str)
    arg2 = self.get_argument(context, index=1, default='', cls=str)
    return arg2 in arg1


@method(function('concat', nargs=(2, None),
                 sequence_types=('xs:anyAtomicType?', 'xs:anyAtomicType?', 'xs:string')))
def evaluate__concat(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    return ''.join(
        self.string_value(self.get_argument(context, index=k)) for k in range(len(self))
    )


@method(function('string-length', nargs=(0, 1),
                 sequence_types=('xs:string?', 'xs:integer')))
def evaluate__string_length(self: XPathFunction, context: ta.ContextType = None) -> int:
    if self.context is not None:
        context = self.context

    if self:
        return len(self.get_argument(context, default_to_context=True, default='', cls=str))
    elif context is None:
        raise self.missing_context()
    else:
        return len(self.string_value(context.item))


@method(function('normalize-space', nargs=(0, 1),
                 sequence_types=('xs:string?', 'xs:string')))
def evaluate__normalize_space(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    if self.parser.version == '1.0' or not self:
        arg = self.string_value(self.get_argument(context, default_to_context=True, default=''))
    else:
        arg = self.get_argument(context, default_to_context=True, default='', cls=str)
    return ' '.join(arg.strip().split())


@method(function('starts-with', nargs=2,
                 sequence_types=('xs:string?', 'xs:string?', 'xs:boolean')))
def evaluate__starts_with(self: XPathFunction, context: ta.ContextType = None) -> bool:
    if self.context is not None:
        context = self.context

    arg1: str = self.get_argument(context, default='', cls=str)
    arg2: str = self.get_argument(context, index=1, default='', cls=str)
    return arg1.startswith(arg2)


@method(function('translate', nargs=3,
                 sequence_types=('xs:string?', 'xs:string', 'xs:string', 'xs:string')))
def evaluate__translate(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    arg: str = self.get_argument(context, default='', cls=str)

    map_string: str = self.get_argument(context, index=1, cls=str)
    if map_string is None:
        message = "the 2nd argument of fn:translate() cannot be the empty sequence"
        raise self.error('XPTY0004', message)

    trans_string: str = self.get_argument(context, index=2, cls=str)
    if trans_string is None:
        message = "the 3rd argument of fn:translate() cannot be the empty sequence"
        raise self.error('XPTY0004', message)

    if len(map_string) == len(trans_string):
        return arg.translate(str.maketrans(map_string, trans_string))
    elif len(map_string) > len(trans_string):
        k = len(trans_string)
        return arg.translate(str.maketrans(map_string[:k], trans_string, map_string[k:]))
    else:
        return arg.translate(str.maketrans(map_string, trans_string[:len(map_string)]))


@method(function('substring', nargs=(2, 3),
                 sequence_types=('xs:string?', 'xs:double', 'xs:double', 'xs:string')))
def evaluate__substring(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    item: str = self.get_argument(context, default='', cls=str)
    try:
        start = self.get_argument(context, index=1, required=True)
        if math.isnan(start) or math.isinf(start):
            return ''
    except TypeError:
        if isinstance(context, XPathSchemaContext):
            start = 0
        else:
            raise self.error('FORG0006', "the second argument must be xs:numeric") from None
    else:
        start = int(round(start)) - 1

    if len(self) == 2:
        return item[max(start, 0):]
    else:
        try:
            length = self.get_argument(context, index=2, required=True)
            if math.isnan(length) or length <= 0:
                return ''
        except TypeError:
            if isinstance(context, XPathSchemaContext):
                length = len(item)
            else:
                raise self.error('FORG0006', "the third argument must be xs:numeric") from None

        if math.isinf(length):
            return item[max(start, 0):]
        else:
            stop = start + int(round(length))
            return item[slice(max(start, 0), max(stop, 0))]


@method(function('substring-before', nargs=2,
                 sequence_types=('xs:string?', 'xs:string?', 'xs:string')))
@method(function('substring-after', nargs=2,
                 sequence_types=('xs:string?', 'xs:string?', 'xs:string')))
def evaluate__substring_before_or_after_functions(
        self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    arg1: str = self.get_argument(context, default='', cls=str)
    arg2: str = self.get_argument(context, index=1, default='', cls=str)

    index = arg1.find(arg2)
    if index < 0:
        return ''
    if self.symbol == 'substring-before':
        return arg1[:index]
    else:
        return arg1[index + len(arg2):]


###
# Boolean functions
@method(function('boolean', nargs=1,
                 sequence_types=('item()*', 'xs:boolean')))
def evaluate__boolean(self: XPathFunction, context: ta.ContextType = None) -> bool:
    return self.boolean_value(self[0].select(self.context or context))


@method(function('not', nargs=1, sequence_types=('item()*', 'xs:boolean')))
def evaluate__not(self: XPathFunction, context: ta.ContextType = None) -> bool:
    return not self.boolean_value(self[0].select(self.context or context))


@method(function('true', nargs=0, sequence_types=('xs:boolean',)))
def evaluate__true(self: XPathFunction, context: ta.ContextType = None) -> bool:
    return True


@method(function('false', nargs=0, sequence_types=('xs:boolean',)))
def evaluate__false(self: XPathFunction, context: ta.ContextType = None) -> bool:
    return False


@method(function('lang', nargs=1,
                 sequence_types=('xs:string?', 'xs:boolean')))
def evaluate__lang(self: XPathFunction, context: ta.ContextType = None) -> bool:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    if not isinstance(context.item, EtreeElementNode):
        return False
    else:
        try:
            attr = context.item.value.attrib[XML_LANG]
        except KeyError:
            for e in context.iter_ancestors():
                if isinstance(e, EtreeElementNode) and XML_LANG in e.value.attrib:
                    lang = e.value.attrib[XML_LANG]
                    if not isinstance(lang, str):
                        return False
                    break
            else:
                return False
        else:
            if not isinstance(attr, str):
                return False
            lang = attr.strip()

        if '-' in lang:
            lang, _ = lang.split('-')

        value = self[0].evaluate()
        if not isinstance(value, str):
            return False
        return lang.lower() == value.lower()


###
# Number functions
@method(function('number', nargs=(0, 1), sequence_types=('xs:anyAtomicType?', 'xs:double')))
def evaluate__number(self: XPathFunction, context: ta.ContextType = None) -> float:
    arg = self.get_argument(self.context or context, default_to_context=True)
    return self.number_value(arg)


@method(function('sum', nargs=(1, 2),
                 sequence_types=('xs:anyAtomicType*', 'xs:anyAtomicType?', 'xs:anyAtomicType?')))
def evaluate__sum(self: XPathFunction, context: ta.ContextType = None) -> ta.OneAtomicOrEmpty:
    if self.context is not None:
        context = self.context

    xsd_version = self.parser.xsd_version
    values: list[Any]
    try:
        values = [get_double(self.string_value(x), xsd_version)
                  if isinstance(x, XPathNode) else x
                  for x in self[0].select_flatten(context)]
    except (TypeError, ValueError):
        if self.parser.version == '1.0':
            return math.nan
        elif isinstance(context, XPathSchemaContext):
            return []
        raise self.error('FORG0006') from None

    if not values:
        zero = 0 if len(self) == 1 else self.get_argument(context, index=1)
        return [] if zero is None else zero

    if all(isinstance(x, (decimal.Decimal, int)) for x in values):
        result = sum(values) if len(values) > 1 else values[0]
    elif all(isinstance(x, DayTimeDuration) for x in values) or \
            all(isinstance(x, YearMonthDuration) for x in values):
        result = sum(values[1:], start=values[0])
    elif any(isinstance(x, Duration) for x in values):
        raise self.error('FORG0006', 'invalid sum of duration values')
    elif any(isinstance(x, (StringProxy, AnyURI)) for x in values):
        raise self.error('FORG0006', 'cannot apply fn:sum() to string-based types')
    elif any(isinstance(x, float) and math.isnan(x) for x in values):
        return math.nan
    elif all(isinstance(x, Float) for x in values):
        result = sum(values)
    else:
        try:
            result = sum(self.number_value(x) for x in values)
        except TypeError:
            if self.parser.version == '1.0':
                return math.nan
            elif isinstance(context, XPathSchemaContext):
                return []
            raise self.error('FORG0006') from None

    assert isinstance(result, AnyAtomicType)
    return result


@method(function('ceiling', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
@method(function('floor', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
def evaluate__ceiling_and_floor_functions(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneNumericOrEmpty:
    if self.context is not None:
        context = self.context

    arg = self.get_argument(context)
    if arg is None:
        return math.nan if self.parser.version == '1.0' else []
    elif isinstance(arg, XPathNode) or self.parser.compatibility_mode:
        arg = self.number_value(arg)

    try:
        if math.isnan(arg) or math.isinf(arg):
            assert isinstance(arg, (int, float, decimal.Decimal))
            return arg

        assert isinstance(arg, (int, float, decimal.Decimal))
        if self.symbol == 'floor':
            return type(arg)(math.floor(arg))
        else:
            return type(arg)(math.ceil(arg))
    except TypeError as err:
        if isinstance(context, XPathSchemaContext):
            return []
        elif isinstance(arg, str):
            raise self.error('XPTY0004', err) from None
        raise self.error('FORG0006', err) from None


@method(function('round', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
def evaluate__round(self: XPathFunction, context: ta.ContextType = None) -> ta.OneNumericOrEmpty:
    if self.context is not None:
        context = self.context

    arg = self.get_argument(context)
    if arg is None:
        return math.nan if self.parser.version == '1.0' else []
    elif isinstance(arg, XPathNode) or self.parser.compatibility_mode:
        arg = self.number_value(arg)

    if isinstance(arg, float) and (math.isnan(arg) or math.isinf(arg)):
        return arg

    try:
        number = decimal.Decimal(arg)
        assert isinstance(arg, (int, float, decimal.Decimal))
        if number > 0:
            return type(arg)(number.quantize(decimal.Decimal('1'), rounding='ROUND_HALF_UP'))
        else:
            return type(arg)(number.quantize(decimal.Decimal('1'), rounding='ROUND_HALF_DOWN'))
    except TypeError as err:
        if isinstance(context, XPathSchemaContext):
            return []
        raise self.error('FORG0006', err) from None
    except decimal.InvalidOperation:
        if not isinstance(arg, str):
            assert isinstance(arg, (int, float, decimal.Decimal))
            return round(arg)
        elif isinstance(context, XPathSchemaContext):
            return []
        raise self.error('XPTY0004') from None
    except decimal.DecimalException as err:
        if isinstance(context, XPathSchemaContext):
            return []
        raise self.error('FOCA0002', err) from None

# XPath 1.0 definitions continue into module xpath1_axes



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
__version__ = '5.1.1'
__author__ = "Davide Brunato"
__contact__ = "brunato@sissa.it"
__copyright__ = "Copyright 2018-2026, SISSA"
__license__ = "MIT"
__status__ = "Production/Stable"

# Imports here are considered as stable API, other internal calls may change.

from . import datatypes  # XSD datatypes
from . import etree      # Safe parser and helper functions for ElementTree
from . import protocols  # Protocols for type annotations

from .exceptions import ElementPathError, MissingContextError, ElementPathKeyError, \
    ElementPathZeroDivisionError, ElementPathNameError, ElementPathOverflowError, \
    ElementPathRuntimeError, ElementPathSyntaxError, ElementPathTypeError, \
    ElementPathValueError, ElementPathLocaleError, UnsupportedFeatureError

from .xpath_context import XPathContext, XPathSchemaContext
from .xpath_nodes import XPathNode, AttributeNode, NamespaceNode, CommentNode, \
    ProcessingInstructionNode, TextNode, ElementNode, LazyElementNode, \
    SchemaElementNode, DocumentNode
from .tree_builders import get_node_tree, build_node_tree, build_lxml_node_tree, \
    build_schema_node_tree
from .xpath_tokens import XPathToken, XPathFunction
from .xpath1 import XPath1Parser
from .xpath2 import XPath2Parser
from .xpath_selectors import select, iter_select, Selector
from .schema_proxy import AbstractSchemaProxy
from .regex import RegexError, translate_pattern, install_unicode_data, unicode_version

__all__ = ['datatypes', 'protocols', 'etree', 'ElementPathError', 'MissingContextError',
           'UnsupportedFeatureError', 'ElementPathKeyError',  'ElementPathLocaleError',
           'ElementPathZeroDivisionError', 'ElementPathNameError',
           'ElementPathOverflowError', 'ElementPathRuntimeError',
           'ElementPathSyntaxError', 'ElementPathTypeError', 'ElementPathValueError',
           'XPathNode', 'AttributeNode', 'NamespaceNode', 'CommentNode',
           'ProcessingInstructionNode', 'TextNode', 'ElementNode',
           'LazyElementNode', 'SchemaElementNode', 'DocumentNode', 'get_node_tree',
           'build_node_tree', 'build_lxml_node_tree', 'build_schema_node_tree',
           'XPathContext', 'XPathSchemaContext', 'XPathToken',
           'XPathFunction', 'XPath1Parser', 'XPath2Parser', 'select', 'iter_select',
           'Selector', 'AbstractSchemaProxy', 'RegexError', 'translate_pattern',
           'install_unicode_data', 'unicode_version']



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module defines common definitions and helper functions for regex subpackage.
"""
from collections.abc import Iterable, Iterator
from typing import Union

CHARACTER_CLASS_ESCAPED: set[int] = {ord(c) for c in r'-|.^?*+{}()[]\\'}
"""Code Points of escaped chars in a character class."""

CodePoint = Union[int, tuple[int, int]]


class RegexError(Exception):
    """
    Error in a regular expression or in a character class specification.
    This exception is derived from `Exception` base class and is raised
    only by the regex subpackage.
    """


def code_point_order(cp: CodePoint) -> int:
    """Ordering function for code points."""
    return cp if isinstance(cp, int) else cp[0]


def code_point_reverse_order(cp: CodePoint) -> int:
    """Reverse ordering function for code points."""
    return cp if isinstance(cp, int) else cp[1] - 1


def iter_code_points(codepoints: Iterable[CodePoint], reverse: bool = False) \
        -> Iterator[CodePoint]:
    """
    Iterates a code points sequence. Three or more consecutive code points are merged in a range.

    :param codepoints: an iterable with code points and code point ranges.
    :param reverse: if `True` reverses the order of the sequence.
    :return: yields code points or code point ranges.
    """
    start_cp = end_cp = 0
    if reverse:
        codepoints = sorted(codepoints, key=code_point_reverse_order, reverse=True)
    else:
        codepoints = sorted(codepoints, key=code_point_order)

    for cp in codepoints:
        if isinstance(cp, int):
            cp0 = cp
            cp1 = cp + 1
        else:
            cp0, cp1 = cp

        if not end_cp:
            start_cp = cp0
            end_cp = cp1
            continue
        elif reverse:
            if start_cp <= cp1:
                if start_cp > cp0:
                    start_cp = cp0
                continue
        elif end_cp >= cp0:
            if end_cp < cp1:
                end_cp = cp1
            continue

        if end_cp > start_cp + 1:
            yield start_cp, end_cp
        else:
            yield start_cp

        start_cp = cp0
        end_cp = cp1
    else:
        if end_cp:
            if end_cp > start_cp + 1:
                yield start_cp, end_cp
            else:
                yield start_cp


def code_point_repr(cp: CodePoint) -> str:
    """
    Returns the string representation of a code point.

    :param cp: an integer or a tuple with at least two integers. \
    Values must be in interval [0, sys.maxunicode].
    """
    if isinstance(cp, int):
        if cp in CHARACTER_CLASS_ESCAPED:
            return r'\%s' % chr(cp)
        return chr(cp)

    if cp[0] in CHARACTER_CLASS_ESCAPED:
        start_char = r'\%s' % chr(cp[0])
    else:
        start_char = chr(cp[0])

    end_cp = cp[1] - 1  # Character ranges include the right bound
    if end_cp in CHARACTER_CLASS_ESCAPED:
        end_char = r'\%s' % chr(end_cp)
    else:
        end_char = chr(end_cp)

    if end_cp > cp[0] + 1:
        return '%s-%s' % (start_char, end_char)
    else:
        return start_char + end_char


def iterparse_character_subset(s: str, expand_ranges: bool = False) -> Iterator[CodePoint]:
    """
    Parses a regex character subset, generating a sequence of code points
    and code points ranges. An unescaped hyphen (-) that is not at the
    start or at the end is interpreted as range specifier.

    :param s: a string representing the character subset.
    :param expand_ranges: if set to `True` then expands character ranges.
    :return: yields integers or couples of integers.
    """
    escaped = False
    on_range = False
    char = ''
    length = len(s)
    subset_index_iterator = iter(range(len(s)))
    for k in subset_index_iterator:
        if k == 0:
            char = s[0]
            if char == '\\':
                escaped = True
            elif char in r'[]' and length > 1:
                raise RegexError("bad character %r at position 0" % char)
            elif expand_ranges:
                yield ord(char)
            elif length <= 2 or s[1] != '-':
                yield ord(char)
        elif s[k] == '-':
            if escaped or (k == length - 1):
                char = s[k]
                yield ord(char)
                escaped = False
            elif on_range:
                char = s[k]
                yield ord(char)
                on_range = False
            else:
                # Parse character range
                on_range = True
                k = next(subset_index_iterator)
                end_char = s[k]
                if end_char == '\\' and (k < length - 1):
                    if s[k + 1] in r'-|.^?*+{}()[]':
                        k = next(subset_index_iterator)
                        end_char = s[k]
                    elif s[k + 1] in r'sSdDiIcCwWpP':
                        msg = "bad character range '%s-\\%s' at position %d: %r"
                        raise RegexError(msg % (char, s[k + 1], k - 2, s))

                if ord(char) > ord(end_char):
                    msg = "bad character range '%s-%s' at position %d: %r"
                    raise RegexError(msg % (char, end_char, k - 2, s))
                elif expand_ranges:
                    yield from range(ord(char) + 1, ord(end_char) + 1)
                else:
                    yield ord(char), ord(end_char) + 1

        elif s[k] in r'|.^?*+{}()':
            if escaped:
                escaped = False
            on_range = False
            char = s[k]
            yield ord(char)
        elif s[k] in r'[]':
            if not escaped and length > 1:
                raise RegexError("bad character %r at position %d" % (s[k], k))
            escaped = on_range = False
            char = s[k]
            if k >= length - 2 or s[k + 1] != '-':
                yield ord(char)
        elif s[k] == '\\':
            if escaped:
                escaped = on_range = False
                char = '\\'
                yield ord(char)
            else:
                escaped = True
        else:
            if escaped:
                escaped = False
                yield ord('\\')
            on_range = False
            char = s[k]
            if k >= length - 2 or s[k + 1] != '-':
                yield ord(char)
    if escaped:
        yield ord('\\')



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import datetime
import importlib
from collections.abc import Iterator, Sequence, Callable
from functools import cached_property
from types import ModuleType
from typing import cast, Any, Optional, Union

import elementpath.aliases as ta

from elementpath.protocols import ElementProtocol, DocumentProtocol
from elementpath.exceptions import ElementPathTypeError
from elementpath.tdop import Token
from elementpath.sequences import XSequence
from elementpath.datatypes import AnyAtomicType, Timezone, Language
from elementpath.etree import is_etree_element, is_etree_element_instance, is_etree_document
from elementpath.xpath_nodes import XPathNode, AttributeNode, NamespaceNode, \
    CommentNode, ProcessingInstructionNode, ElementNode, DocumentNode
from elementpath.tree_builders import get_node_tree

__all__ = ['XPathContext', 'XPathSchemaContext']


class XPathContext:
    """
    The XPath dynamic context. The static context is provided by the parser.

    Usually the dynamic context instances are created providing only the root element.
    Variable values argument is needed if the XPath expression refers to in-scope variables.
    The other optional arguments are needed only if a specific position on the context is
    required, but have to be used with the knowledge of what is their meaning.

    :param root: the root of the XML document, usually an ElementTree instance or an \
    Element. A schema or a schema element can also be provided, or an already built \
    node tree. For default is `None`, in which case no XML root is set, and you have \
    to provide an *item* argument.
    :param namespaces: a dictionary with mapping from namespace prefixes into URIs, \
    used when namespace information is not available within document and element nodes. \
    This can be useful when the dynamic context has additional namespaces and root \
    is an Element or an ElementTree instance of the standard library.
    :param uri: an optional URI associated with the root element or the document.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    In this case the dummy document value is not included in results. For default the \
    root node kind is preserved.
    :param item: the context item. A `None` value means that the context is positioned on \
    the document node.
    :param position: the current position of the node within the input sequence.
    :param size: the number of items in the input sequence.
    :param axis: the active axis. Used to choose when apply the default axis ('child' axis).
    :param schema: an optional schema proxy instance to be applied on XDM root or item.
    :param variables: dictionary of context variables that maps a QName to a value.
    :param current_dt: current dateTime of the implementation, including explicit timezone.
    :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
    not have a timezone.
    :param documents: available documents. This is a mapping of absolute URI \
    strings into document nodes. Used by the function fn:doc.
    :param collections: available collections. This is a mapping of absolute URI \
    strings onto sequences of nodes. Used by the XPath 2.0+ function fn:collection.
    :param default_collection: this is the sequence of nodes used when fn:collection \
    is called with no arguments.
    :param text_resources: available text resources. This is a mapping of absolute URI strings \
    onto text resources. Used by XPath 3.0+ function fn:unparsed-text/fn:unparsed-text-lines.
    :param resource_collections: available URI collections. This is a mapping of absolute \
    URI strings to sequence of URIs. Used by the XPath 3.0+ function fn:uri-collection.
    :param default_resource_collection: this is the sequence of URIs used when \
    fn:uri-collection is called with no arguments.
    :param allow_environment: defines if the access to system environment is allowed, \
    for default is `False`. Used by the XPath 3.0+ functions fn:environment-variable \
    and fn:available-environment-variables.
    """
    _etree: Optional[ModuleType] = None
    _schema: Optional[ta.SchemaProxyType] = None
    root: Optional[ta.RootNodeType]
    document: DocumentNode | None
    item: ta.ItemType

    documents: dict[str, DocumentNode] | None = None
    collections: dict[str, list[XPathNode]] | None = None
    default_collection: list[XPathNode] | None = None

    __slots__ = ('document', 'root', 'item', 'namespaces', 'size',
                 'position', 'variables', 'axis', '__dict__')

    def __init__(self,
                 root: Optional[ta.RootArgType] = None,
                 namespaces: Optional[ta.NamespacesType] = None,
                 uri: str | None = None,
                 fragment: bool | None = None,
                 item: Optional[ta.ItemArgType] = None,
                 position: int = 1,
                 size: int = 1,
                 axis: str | None = None,
                 schema: Optional[ta.SchemaProxyType] = None,
                 variables: dict[str, ta.VariableValueType] | None = None,
                 current_dt: datetime.datetime | None = None,
                 timezone: Union[str, Timezone] | None = None,
                 documents: dict[str, ta.RootArgType] | None = None,
                 collections: dict[str, ta.CollectionArgType] | None = None,
                 default_collection: ta.CollectionArgType = None,
                 text_resources: dict[str, str] | None = None,
                 resource_collections: dict[str, list[str]] | None = None,
                 default_resource_collection: str | None = None,
                 allow_environment: bool = False,
                 default_language: str | None = None,
                 default_calendar: str | None = None,
                 default_place: str | None = None) -> None:

        if namespaces:
            self.namespaces = {k: v for k, v in namespaces.items()}
        else:
            self.namespaces = {}

        if root is not None:
            self.root = get_node_tree(root, self.namespaces, uri, fragment)
            if item is not None:
                self.item = self.get_context_item(item, self.namespaces)
            else:
                self.item = self.root

        elif item is not None:
            self.root = None
            self.item = self.get_context_item(item, self.namespaces, uri, fragment)
        else:
            raise ElementPathTypeError("Missing both the root node and the context item!")

        if isinstance(self.root, DocumentNode):
            self.document = self.root
        elif fragment is None and \
                isinstance(self.root, ElementNode) and \
                is_etree_element_instance(self.root.value):
            # Creates a dummy document that will be not included in results
            self.document = self.root.get_document_node(as_parent=False)
        else:
            self.document = None

        self.position = position
        self.size = size
        self.axis = axis

        if timezone is None or isinstance(timezone, Timezone):
            self.timezone = timezone
        else:
            self.timezone = Timezone.fromstring(timezone)
        self.current_dt = current_dt or datetime.datetime.now(tz=self.timezone)

        if documents is not None:
            # Assume that are all documents because type checking is done by fn:doc().
            self.documents = {
                k: cast(DocumentNode, get_node_tree(v, self.namespaces, k))
                if v is not None else v for k, v in documents.items()
            }

        if schema is not None:
            self.schema = schema

        self.variables = dict[str, ta.ValueType]()
        if variables is not None:
            for varname, value in variables.items():
                self.variables[varname] = self.get_value(value, self.namespaces)

        if collections is not None:
            self.collections = {k: self.get_collection(v) for k, v in collections.items()}

        if default_collection is not None:
            self.default_collection = self.get_collection(default_collection)

        self.text_resources = text_resources if text_resources is not None else {}
        self.resource_collections = resource_collections
        self.default_resource_collection = default_resource_collection
        self.allow_environment = allow_environment
        self.default_language = None if default_language is None else Language(default_language)
        self.default_calendar = default_calendar
        self.default_place = default_place

    def __repr__(self) -> str:
        if self.root is not None:
            return f'{self.__class__.__name__}(root={self.root.value})'
        elif isinstance(self.item, XPathNode):
            return f'{self.__class__.__name__}(item={self.item.value})'
        else:
            return f'{self.__class__.__name__}(item={self.item!r})'

    def __copy__(self) -> 'XPathContext':
        obj: XPathContext = object.__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        obj.document = self.document
        obj.root = self.root
        obj.item = self.item
        obj.size = self.size
        obj.position = self.position
        obj.axis = None
        obj.namespaces = self.namespaces
        obj.variables = self.variables
        return obj

    @cached_property
    def etree(self) -> ModuleType:
        if isinstance(self.root, (DocumentNode, ElementNode)):
            module_name = self.root.value.__class__.__module__
        elif isinstance(self.item, (DocumentNode, ElementNode, CommentNode,
                                    ProcessingInstructionNode)):
            module_name = self.item.value.__class__.__module__
        else:
            module_name = 'xml.etree.ElementTree'

        if module_name in ('lxml.etree', 'lxml.html'):
            return importlib.import_module('lxml.etree')
        else:
            return importlib.import_module('xml.etree.ElementTree')

    @property
    def schema(self) -> Optional[ta.SchemaProxyType]:
        return self._schema

    @schema.setter
    def schema(self, schema: Optional[ta.SchemaProxyType]) -> None:
        self._schema = schema
        if schema is None:
            if self.root is not None:
                self.root.clear_types()
            elif isinstance(self.item, XPathNode):
                self.item.clear_types()
        elif hasattr(schema, 'is_assertion_based'):
            if self.root is not None:
                self.root.clear_types()
                self.root.apply_schema(schema)
            elif isinstance(self.item, XPathNode):
                self.item.clear_types()
                self.item.apply_schema(schema)
        else:
            msg = f"{schema!r} is not an instance of AbstractSchemaProxy"
            raise ElementPathTypeError(msg)

    def get_root(self, node: Any) -> ta.RootNodeType | None:
        if isinstance(self.root, (DocumentNode, ElementNode)):
            if any(node is x for x in self.root.iter_lazy()):
                return self.root

        if self.documents is not None:
            for uri, doc in self.documents.items():
                if doc is not None and any(node is x for x in doc.iter_lazy()):
                    return doc

        return None

    def is_document(self) -> bool:
        return isinstance(self.document, DocumentNode)

    def is_fragment(self) -> bool:
        return self.document is None and self.root is not None

    def is_rooted_subtree(self) -> bool:
        return self.root is not None and isinstance(self.root.parent, ElementNode)

    def is_principal_node_kind(self) -> bool:
        if self.axis == 'attribute':
            return isinstance(self.item, AttributeNode)
        elif self.axis == 'namespace':
            return isinstance(self.item, NamespaceNode)
        else:
            return isinstance(self.item, ElementNode)

    def get_context_item(self, item: ta.ItemArgType,
                         namespaces: ta.NamespacesType | None = None,
                         uri: Optional[str] = None,
                         fragment: Optional[bool] = None) -> ta.ItemType:
        """
        Checks the item and returns an item suitable for XPath processing.
        For XML trees and elements try a match with an existing node in the
        context. If it fails then builds a new node using also the provided
        optional arguments.
        """
        if isinstance(item, (XPathNode, AnyAtomicType)):
            return item
        elif is_etree_document(item):
            if self.root is not None and item is self.root.value:
                return self.root

            if self.documents:
                for doc in self.documents.values():
                    if doc is not None and item is doc.value:
                        return doc

        elif is_etree_element(item):
            try:
                return self.root.elements[item]  # type: ignore[index,union-attr]
            except (TypeError, KeyError, AttributeError):
                pass

            if self.documents:
                for doc in self.documents.values():
                    if doc is not None and doc.elements is not None and item in doc.elements:
                        return doc.elements[item]

            if callable(item.tag):  # type: ignore[union-attr]
                if item.tag.__name__ == 'Comment':  # type: ignore[union-attr]
                    return CommentNode(cast(ElementProtocol, item))
                else:
                    return ProcessingInstructionNode(cast(ElementProtocol, item))
        elif not isinstance(item, Token) or not callable(item):
            msg = f"Unexpected type {type(item)} for context item: {item!r}"
            raise ElementPathTypeError(msg)
        else:
            return item

        return get_node_tree(
            root=cast(Union[ElementProtocol, DocumentProtocol], item),
            namespaces=namespaces,
            uri=uri,
            fragment=fragment
        )

    def get_value(self, item: ta.FunctionArgType, *args: Any, **kwargs: Any) -> ta.ValueType:
        if item is None:
            return []
        elif not isinstance(item, (list, tuple, XSequence)):
            return self.get_context_item(item, *args, **kwargs)
        elif not item:
            return []
        else:
            return [self.get_context_item(x, *args, **kwargs) for x in item]

    def get_collection(self, items: ta.CollectionArgType) -> list[XPathNode]:
        if items is None:
            return []
        elif isinstance(items, (list, tuple)):
            return [x for x in map(self.get_context_item, items) if isinstance(x, XPathNode)]
        else:
            item = self.get_context_item(items)
            return [item] if isinstance(item, XPathNode) else []

    def inner_focus_select(self, token: ta.XPathTokenType, predicate: bool = False) \
            -> Iterator[ta.ItemType]:
        return token.select_with_focus(self)

    def iter_product(self, selectors: Sequence[Callable[[Any], Any]],
                     varnames: Optional[Sequence[str]] = None) -> Iterator[Any]:
        """
        Iterator for cartesian products of selectors.

        :param selectors: a sequence of selector generator functions.
        :param varnames: a sequence of variables for storing the generated values.
        """
        if varnames is None:
            varnames = []
        iterators = [x(self) for x in selectors]
        dimension = len(iterators)
        prod = [None] * dimension
        max_index = dimension - 1

        k = 0
        while True:
            for value in iterators[k]:
                try:
                    self.variables[varnames[k]] = value
                except IndexError:
                    pass

                prod[k] = value
                if k == max_index:
                    yield tuple(prod)
                else:
                    k += 1
                break
            else:
                if not k:
                    return
                iterators[k] = selectors[k](self)
                k -= 1

    ##
    # Context item iterators for axis

    def iter_self(self) -> Iterator[ta.ItemType]:
        """Iterator for 'self' axis and '.' shortcut."""
        if self.item is not None:
            status = self.axis
            self.axis = 'self'
            yield self.item
            self.axis = status

    def iter_attributes(self) -> Iterator[AttributeNode]:
        """Iterator for 'attribute' axis and '@' shortcut."""
        status: Any

        if isinstance(self.item, AttributeNode):
            status = self.axis
            self.axis = 'attribute'
            yield self.item
            self.axis = status
            return
        elif isinstance(self.item, ElementNode):
            status = self.item, self.axis
            self.axis = 'attribute'

            for self.item in self.item.attributes:
                yield self.item

            self.item, self.axis = status

    def iter_children_or_self(self) -> Iterator[ta.ItemType]:
        """Iterator for 'child' forward axis and '/' step."""
        if self.item is not None:
            if self.axis is not None:
                yield self.item
            elif isinstance(self.item, (ElementNode, DocumentNode)):
                _status = self.item, self.axis
                self.axis = 'child'

                if self.item is self.document and self.root is not self.document:
                    if self.root is not None:
                        yield self.root
                else:
                    for self.item in self.item:
                        yield self.item

                self.item, self.axis = _status

    def iter_matching_nodes(self, name: str, default_namespace: Optional[str] = None) \
            -> Iterator[Union[AttributeNode, ElementNode]]:
        """
        Iterator for matching elements or attributes. For default uses 'child'
        forward axis if no axis is active, otherwise tests the current item.
        """
        if self.axis is not None:
            if isinstance(self.item, (AttributeNode, ElementNode)):
                if self.item.match_name(name, default_namespace):
                    yield self.item
        elif isinstance(self.item, (ElementNode, DocumentNode)):
            _status = self.item, self.axis
            self.axis = 'child'

            if self.item is self.document and isinstance(self.root, ElementNode):
                if self.root.match_name(name, default_namespace):
                    yield self.root
            else:
                for self.item in self.item:
                    if self.item.match_name(name, default_namespace):
                        assert isinstance(self.item, ElementNode)
                        yield self.item

            self.item, self.axis = _status

    def iter_parent(self) -> Iterator[ta.RootNodeType]:
        """Iterator for 'parent' reverse axis and '..' shortcut."""
        if isinstance(self.item, XPathNode):

            # A stop rule for non-rooted fragments (e.g. root is a schema elements)
            if self.document is not None or self.item is not self.root:
                if self.item.parent is not None:
                    status = self.item, self.axis
                    self.axis = 'parent'

                    self.item = self.item.parent
                    yield self.item

                    self.item, self.axis = status

    def iter_siblings(self, axis: str | None = None) -> Iterator[ta.ChildNodeType]:
        """
        Iterator for 'following-sibling' forward axis and 'preceding-sibling' reverse axis.

        :param axis: the context axis, default is 'following-sibling'.
        """
        if isinstance(self.item, XPathNode):
            if self.document is not None or self.item is not self.root:
                item = self.item

                if item.parent is not None:
                    status = self.item, self.axis
                    self.axis = axis or 'following-sibling'

                    if axis == 'preceding-sibling':
                        for child in item.parent:  # pragma: no cover
                            if child is item:
                                break
                            self.item = child
                            yield child
                    else:
                        follows = False
                        for child in item.parent:
                            if follows:
                                self.item = child
                                yield child
                            elif child is item:
                                follows = True

                    self.item, self.axis = status

    def iter_descendants(self, axis: Optional[str] = None) -> Iterator[Union[None, XPathNode]]:
        """
        Iterator for 'descendant' and 'descendant-or-self' forward axes and '//' shortcut.

        :param axis: the context axis, for default has no explicit axis.
        """
        if isinstance(self.item, (DocumentNode, ElementNode)):
            status = self.item, self.axis
            self.axis = axis

            for self.item in self.item.iter_descendants(with_self=axis != 'descendant'):
                yield self.item

            self.item, self.axis = status

        elif axis != 'descendant' and isinstance(self.item, XPathNode):
            self.axis, axis = axis, self.axis
            yield self.item
            self.axis = axis

    def iter_ancestors(self, axis: Optional[str] = None) -> Iterator[XPathNode]:
        """
        Iterator for 'ancestor' and 'ancestor-or-self' reverse axes.

        :param axis: the context axis, default is 'ancestor'.
        """
        if isinstance(self.item, XPathNode):
            status = self.item, self.axis
            self.axis = axis or 'ancestor'

            ancestors: list[XPathNode] = []
            if axis == 'ancestor-or-self':
                ancestors.append(self.item)

            if self.document is not None or self.item is not self.root:
                parent = self.item.parent
                while parent is not None:
                    ancestors.append(parent)
                    if parent is self.root and self.document is None:
                        break
                    parent = parent.parent

            for self.item in reversed(ancestors):
                yield self.item

            self.item, self.axis = status

    def iter_preceding(self) -> Iterator[Union[DocumentNode, ta.ChildNodeType]]:
        """Iterator for 'preceding' reverse axis."""
        ancestors: set[ta.RootNodeType]
        item: XPathNode

        if isinstance(self.item, XPathNode):
            if self.document is not None or self.item is not self.root:
                item = self.item

                if (root := item.parent) is not None:
                    status = self.item, self.axis
                    self.axis = 'preceding'
                    ancestors = {root}

                    while root.parent is not None:
                        if root is self.root and self.document is None:
                            break
                        root = root.parent
                        ancestors.add(root)

                    for self.item in root.iter_descendants():
                        if self.item is item:
                            break
                        if self.item not in ancestors:
                            yield self.item

                    self.item, self.axis = status

    def iter_followings(self) -> Iterator[ta.ChildNodeType]:
        """Iterator for 'following' forward axis."""
        if isinstance(self.item, ElementNode):
            status = self.item, self.axis
            self.axis = 'following'

            descendants = set(self.item.iter_descendants())
            position = self.item.position

            root = self.item
            while isinstance(root.parent, ElementNode) and root is not self.root:
                root = root.parent

            for item in root.iter_descendants(with_self=False):
                if position < item.position and item not in descendants:
                    self.item = item
                    yield item

            self.item, self.axis = status


class XPathSchemaContext(XPathContext):
    """
    The XPath dynamic context base class for schema bounded parsers. Use this class
    as dynamic context for schema instances in order to perform a schema-based type
    checking during the static analysis phase. Don't use this as dynamic context on
    XML instances.
    """
    root: ElementNode

    @property
    def schema(self) -> Optional[ta.SchemaProxyType]:
        return self._schema

    @schema.setter
    def schema(self, schema: Optional[ta.SchemaProxyType]) -> None:
        self._schema = schema

    def iter_matching_nodes(self, name: str, default_namespace: Optional[str] = None) \
            -> Iterator[Union[AttributeNode, ElementNode]]:
        """
        Iterator for matching elements or attributes. For default uses 'child'
        forward axis if no axis is active, otherwise tests the current item.
        """
        if self.axis is not None:
            if isinstance(self.item, (AttributeNode, ElementNode)):
                if self.item.match_name(name, default_namespace):
                    if not self.item.name:
                        if isinstance(self.item, ElementNode):
                            for element_node in self.root:
                                assert isinstance(element_node, ElementNode)
                                if element_node.match_name(name, default_namespace):
                                    self.item = element_node
                                    break
                        else:
                            for attribute_node in self.root.attributes:
                                if attribute_node.match_name(name, default_namespace):
                                    self.item = attribute_node
                                    break

                    yield self.item

        elif isinstance(self.item, ElementNode):
            _status = self.item, self.axis
            self.axis = 'child'

            for self.item in self.item:
                if self.item.match_name(name, default_namespace):
                    if not self.item.name:
                        for element_node in self.root:
                            if element_node.match_name(name, default_namespace):
                                self.item = element_node
                                break

                    assert isinstance(self.item, ElementNode)
                    yield self.item

            self.item, self.axis = _status



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import math
from collections.abc import KeysView, ValuesView, ItemsView, Iterator
from types import MappingProxyType
from typing import Optional, Union, Any

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathValueError
from elementpath.datatypes import AnyAtomicType
from elementpath.sequences import xlist
from elementpath.helpers import split_function_test
from elementpath.sequence_types import match_sequence_type
from elementpath.xpath_context import XPathSchemaContext
from .functions import XPathFunction


class MapKeysView(KeysView[Optional[ta.AtomicType]]):
    _mapping: MappingProxyType[Optional[ta.AtomicType], ta.ValueType]

    __slots__ = ()

    def __contains__(self, key: object) -> bool:
        if key is None:
            return False
        elif isinstance(key, float) and math.isnan(key):
            return None in self._mapping
        return key in self._mapping

    def __iter__(self) -> Iterator[ta.AtomicType]:
        for k in self._mapping:
            if k is None:
                yield float('nan')
            else:
                yield k


class MapsItemsView(ItemsView[Optional[ta.AtomicType], ta.ValueType]):
    _mapping: MappingProxyType[Optional[ta.AtomicType], ta.ValueType]

    __slots__ = ()

    def __contains__(self, item: Any) -> bool:
        key, value = item
        if key is None:
            return False

        try:
            if isinstance(key, float) and math.isnan(key):
                v = self._mapping[None]
            else:
                v = self._mapping[key]
        except KeyError:
            return False
        else:
            return v is value or v == value

    def __iter__(self) -> Iterator[tuple[ta.AtomicType, ta.ValueType]]:
        for k in self._mapping:
            if k is None:
                yield float('nan'), self._mapping[k]
            else:
                yield k, self._mapping[k]


class XPathMap(XPathFunction):
    """
    A token for processing XPath 3.1+ maps. Map instances have the double role of
    tokens and of dictionaries, depending on the way that are created (using a map
    constructor or a function). The map is fully set after the protected attribute
    _map is evaluated from tokens or initialized from arguments.
    """
    symbol = 'map'
    label = 'map'
    pattern = r'(?<!\$)\bmap(?=\s*(?:\(\:.*\:\))?\s*\{(?!\:))'
    _map: Optional[ta.MapDictType] = None
    _values: list[ta.XPathTokenType]  # a 2nd list of tokens is needed for map's values
    _nan_key: Union[bool, float] = False

    def __init__(self, parser: ta.XPathParserType, items: Optional[Any] = None) -> None:
        super().__init__(parser)
        self._values = []
        if items is not None:
            _items = items.items() if isinstance(items, dict) else items
            _map: ta.MapDictType = {}
            for k, v in _items:
                if k is None:
                    raise self.error('XPTY0004', 'missing key value')
                elif isinstance(k, float) and math.isnan(k):
                    if self._nan_key is False:
                        raise self.error('XQDY0137')
                    self._nan_key, _map[None] = k, v
                    continue
                elif k in _map:
                    raise self.error('XQDY0137')

                if isinstance(v, list):
                    _map[k] = xlist(v)
                else:
                    _map[k] = v

            self._map = _map

    def __repr__(self) -> str:
        if self._map is not None:
            return f'<{self.__class__.__name__} object at {hex(id(self))}>'
        return "<{} object (not evaluated constructor) at {}>".format(
            self.__class__.__name__, hex(id(self))
        )

    def __str__(self) -> str:
        if self._map is None:
            return f'not evaluated map constructor with {len(self._items)} entries'
        return f'map{self._map}'

    def __len__(self) -> int:
        if self._map is None:
            return len(self._items)
        return len(self._map)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, XPathMap):
            if self._map is None or other._map is None:
                raise ElementPathValueError("cannot compare not evaluated maps")
            return self._map == other._map
        return NotImplemented

    def nud(self) -> 'XPathMap':
        self.parser.advance('{')
        del self._items[:]
        if self.parser.next_token.symbol not in ('}', '(end)'):
            while True:
                key = self.parser.expression(5)
                self._items.append(key)
                if self.parser.token.symbol != ':':
                    self.parser.advance(':')
                self._values.append(self.parser.expression(5))

                if self.parser.next_token.symbol != ',':
                    break
                self.parser.advance()

        self.parser.advance('}')
        return self

    @property
    def source(self) -> str:
        if self._map is None:
            items = ', '.join(f'{tk.source}:{tv.source}' for tk, tv in zip(self, self._values))
        else:
            items = ', '.join(f'{k!r}:{v!r}' for k, v in self._map.items())
        return f'map{{{items}}}'

    def evaluate(self, context: ta.ContextType = None) -> 'XPathMap':
        if self._map is not None:
            return self
        return XPathMap(
            parser=self.parser,
            items=(
                (k.get_atomized_operand(context), v.evaluate(context))
                for k, v in zip(self._items, self._values)
            )
        )

    def _evaluate(self, context: ta.ContextType = None) -> ta.MapDictType:
        _map: ta.MapDictType = {}
        nan_key: Union[bool, float] = False

        for key, value in zip(self._items, self._values):
            k = key.get_atomized_operand(context)
            if k is None:
                raise self.error('XPTY0004', 'missing key value')
            elif isinstance(k, float) and math.isnan(k):
                if nan_key is not False:
                    raise self.error('XQDY0137')
                nan_key, _map[None] = k, value.evaluate(context)
                continue
            elif k in _map:
                raise self.error('XQDY0137')

            v = value.evaluate(context)
            if isinstance(v, list):
                _map[k] = xlist(v)
            else:
                _map[k] = v

        self._nan_key = nan_key
        return _map

    def __call__(self, *args: ta.FunctionArgType,
                 context: ta.ContextType = None) -> ta.ValueType:
        if len(args) == 1 and isinstance(args[0], list) and len(args[0]) == 1:
            args = args[0][0],
        if len(args) != 1 or not isinstance(args[0], AnyAtomicType):
            if isinstance(context, XPathSchemaContext):
                return []
            raise self.error('XPST0003', 'exactly one atomic argument is expected')

        _map: ta.MapDictType
        key = args[0]
        if self._map is not None:
            _map = self._map
        else:
            _map = self._evaluate(context)

        try:
            if isinstance(key, float) and math.isnan(key):
                return _map[None]
            else:
                return _map[key]
        except KeyError:
            return []

    def keys(self, context: ta.ContextType = None) -> MapKeysView:
        if self._map is None:
            self._map = self._evaluate(context)
        return MapKeysView(MappingProxyType(self._map))

    def values(self, context: ta.ContextType = None) -> ValuesView[ta.ValueType]:
        if self._map is None:
            self._map = self._evaluate(context)
        return self._map.values()

    def items(self, context: ta.ContextType = None) -> MapsItemsView:
        if self._map is None:
            self._map = self._evaluate(context)
        return MapsItemsView(MappingProxyType(self._map))

    def match_function_test(self, function_test: ta.SequenceTypesType,
                            as_argument: bool = False) -> bool:
        if isinstance(function_test, (list, tuple)):
            sequence_types = function_test
        else:
            sequence_types = split_function_test(function_test)

        if not sequence_types or not sequence_types[-1]:
            return False
        elif sequence_types[0] == '*':
            return True
        elif len(sequence_types) != 2:
            return False

        key_st, value_st = sequence_types
        if key_st.endswith(('+', '*')):
            return False
        elif value_st != 'empty-sequence()' and not value_st.endswith(('?', '*')):
            return False
        else:
            return any(match_sequence_type(k, key_st, self.parser, False) and
                       match_sequence_type(v, value_st, self.parser)
                       for k, v in self.items())



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Sequence
from types import MappingProxyType
from typing import Any, overload, TypeVar

from .any_types import AtomicTypeMeta, AnySimpleType, AnyAtomicType
from .untyped import UntypedAtomic
from .string import NMToken, Idref, Entity

__all__ = ['builtin_list_types', 'ListType', 'NMTokens', 'Idrefs', 'Entities']


_builtin_list_types: 'dict[str, AtomicTypeMeta]' = {}
builtin_list_types = MappingProxyType(_builtin_list_types)
"""Registry of builtin list types by expanded name."""

T = TypeVar('T', bound=AnyAtomicType)


class ListTypeMeta(AtomicTypeMeta):
    types_map = _builtin_list_types


class ListType(Sequence[AnyAtomicType], AnySimpleType, metaclass=ListTypeMeta):
    value: list[AnyAtomicType]
    item_type: type[AnyAtomicType]

    __slots__ = ('value',)

    def __init__(self, value: list[AnyAtomicType]) -> None:
        self.value = value

    @overload
    def __getitem__(self, index: int) -> AnyAtomicType: ...

    @overload
    def __getitem__(self, index: slice) -> list[AnyAtomicType]: ...

    def __getitem__(self, index: int | slice) -> AnyAtomicType | list[AnyAtomicType]:
        return self.value[index]

    def __len__(self) -> int:
        return len(self.value)

    @classmethod
    def validate(cls, obj: object) -> None:
        if isinstance(obj, cls):
            return
        elif not isinstance(obj, list) or \
                any(not isinstance(item, AnyAtomicType) for item in obj):
            raise cls._invalid_type(obj)
        else:
            for item in obj:
                cls.item_type.validate(item)

    @classmethod
    def decode(cls, value: Any) -> list[AnyAtomicType]:
        if isinstance(value, UntypedAtomic):
            values = value.value.split() or ['']
        elif isinstance(value, str):
            values = value.split() or ['']
        elif isinstance(value, list):
            values = value
        else:
            raise cls._invalid_type(value)

        try:
            return [cls.item_type(x) for x in values]
        except ValueError:
            raise cls._invalid_value(value)


###
# Builtin list types

class NMTokens(ListType):
    name = 'NMTOKENS'
    items: list[NMToken]
    item_type: type[NMToken] = NMToken


class Idrefs(ListType):
    name = 'IDREFS'
    items: list[Idref]
    item_type: type[Idref] = Idref


class Entities(ListType):
    name = 'ENTITIES'
    items: list[Entity]
    item_type: type[Entity] = Entity



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import importlib
from collections import defaultdict, deque
from collections.abc import Iterator
from urllib.parse import urljoin
from typing import cast, Any, Optional
from xml.etree import ElementTree

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathRuntimeError, \
    ElementPathValueError, ElementPathKeyError
from elementpath.datatypes import UntypedAtomic, AnyURI, QName
from elementpath.namespaces import XML_NAMESPACE, XML_BASE, XSI_NIL, \
    XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSI_TYPE, \
    XML_ID, XSD_IDREF, XSD_IDREFS, XSD_UNTYPED, XSD_UNTYPED_ATOMIC, \
    XPATH_FUNCTIONS_NAMESPACE, get_expanded_name
from elementpath.protocols import ElementProtocol, XsdElementProtocol, \
    XsdAttributeProtocol, XsdTypeProtocol, DocumentType, ElementType, \
    SchemaElemType, CommentType, ProcessingInstructionType
from elementpath.helpers import match_wildcard, is_absolute_uri
from elementpath.decoder import get_atomic_sequence
from elementpath.etree import etree_iter_strings, is_etree_element_instance

__all__ = ['XPathNodeTree', 'XPathNode', 'NamespaceNode', 'AttributeNode', 'TextAttributeNode',
           'SchemaAttributeNode', 'TextNode', 'CommentNode', 'ProcessingInstructionNode',
           'ElementNode', 'EtreeElementNode', 'LazyElementNode', 'SchemaElementNode',
           'DocumentNode', 'EtreeDocumentNode']

_EMPTY_NAME_PATH = f'*[Q{{{XPATH_FUNCTIONS_NAMESPACE}}}local-name()=""]'

_XSD_SPECIAL_TYPES = frozenset((XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE))


class XPathNodeTree:
    """
    Status of the node tree structure, shared between nodes.
    """
    root: ta.ParentNodeType
    elements: ta.ElementMapType
    namespaces: ta.NamespacesType
    schema: Optional[ta.SchemaProxyType]
    uri: str | None

    __slots__ = ('root_node', 'uri', 'namespaces', 'ns_offset', 'elements', 'schema', 'position')

    def __init__(self, root: ta.ParentNodeType,
                 uri: str | None = None,
                 namespaces: Optional[ta.NamespacesType] = None,
                 elements: Optional[ta.ElementMapType] = None,
                 schema: Optional[ta.SchemaProxyType] = None) -> None:

        self.root_node = root
        self.namespaces = {} if namespaces is None else namespaces
        self.uri = uri
        self.elements = {} if elements is None else elements
        self.schema = schema


###
# XQuery and XPath Data Model: https://www.w3.org/TR/xpath-datamodel/
#
# Note: in this implementation empty sequence return value is replaced by None.
#
# XPath has seven kinds of nodes:
#
#  element, attribute, text, namespace, processing-instruction, comment, document
###
class XPathNode:
    """
    The base class of all XPath nodes. In the base class and in other intermediate
    derivation string and typed values are not implemented. Use these classes only
    for type checking and for wrapping other types in a custom XPath node types.
    """
    __slots__ = ('name', 'value', 'parent', 'position')

    ###
    # XDM accessors

    @property
    def attributes(self) -> list['AttributeNode'] | None:
        return None

    @property
    def base_uri(self) -> str | None:
        return self.parent.base_uri if self.parent is not None else None

    children: list[ta.ChildNodeType] | None

    @property
    def document_uri(self) -> str | None:
        return None

    @property
    def is_id(self) -> bool | None:
        return None

    @property
    def is_idrefs(self) -> bool | None:
        return None

    @property
    def namespace_nodes(self) -> list['NamespaceNode'] | None:
        return None

    @property
    def nilled(self) -> bool | None:
        return None

    @property
    def node_kind(self) -> str:
        raise NotImplementedError()

    @property
    def node_name(self) -> QName | None:
        name: str | None = getattr(self, 'name', None)
        if name is None:
            return None
        elif not name.startswith('{'):
            return QName(None, name)

        try:
            namespace, local = name[1:].split('}')
        except ValueError:
            raise ElementPathValueError(f'invalid name format for {self!r}')
        else:
            if namespace == XML_NAMESPACE:
                return QName(namespace, f'xml:{local}')

        if isinstance(self, ElementNode):
            nsmap = self.nsmap
        elif isinstance(self.parent, ElementNode):
            nsmap = self.parent.nsmap
        else:
            nsmap = {}

        for prefix, ns in nsmap.items():
            if namespace == ns:
                if not prefix:
                    return QName(namespace, local)
                return QName(namespace, f"{prefix}:{local}")
        raise ElementPathKeyError(f'missing namespace prefix mapping in {self!r}')

    parent: ta.ParentNodeType | None

    @property
    def type_name(self) -> str | None:
        return None

    @property
    def string_value(self) -> str:
        raise NotImplementedError()

    @property
    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
        raise NotImplementedError()

    @staticmethod
    def unparsed_entity_public_id(name: str) -> str | None:
        return None

    @staticmethod
    def unparsed_entity_system_id(name: str) -> AnyURI | None:
        return None

    ###
    # Other properties and methods

    name: str | None  # node name
    value: object        # the object wrapped in the node
    position: int        # position of the node, for document total order

    @property
    def obj(self) -> Any:
        """Access to wrapped object using the old API."""
        return self.value

    @property
    def compat_string_value(self) -> str:
        return self.string_value

    @property
    def root_node(self) -> 'XPathNode':
        return self if self.parent is None else self.parent.tree.root_node

    @property
    def path(self) -> str:
        """Returns the node path in XPath 3.0+ format."""
        return ''

    @property
    def extended_path(self) -> str:
        """Returns the node path in extended format."""
        return self.path.replace('Q{}', '').replace('Q{', '{')

    @property
    def qname_path(self) -> str:
        """Returns the node path with names in prefixed QName format."""
        path = self.path

        if isinstance(self, ElementNode):
            for prefix, namespace in self.nsmap.items():
                path = path.replace(f'Q{{{namespace}}}', f'{prefix}:')
        elif isinstance(self.parent, ElementNode):
            for prefix, namespace in self.parent.nsmap.items():
                path = path.replace(f'Q{{{namespace}}}', f'{prefix}:')

        path = path.replace('Q{}', '')
        if 'Q{' not in path:
            return path
        raise ElementPathKeyError(f'missing namespace prefix mapping in {path}')

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        raise NotImplementedError()

    @property
    def is_schema_node(self) -> bool | None:
        return None

    @property
    def is_typed(self) -> bool | None:
        return None

    @property
    def is_extended(self) -> bool | None:
        return None

    @property
    def is_list(self) -> bool | None:
        return None

    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
        """Set XSD types for elements and attribute nodes from schema proxy instance."""
        if self.parent is not None:
            self.parent.apply_schema(schema)

    def clear_types(self) -> None:
        """Clear XSD types for elements and attribute nodes."""
        if self.parent is not None:
            self.parent.clear_types()

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        """
        Returns `True` if the argument is matching the name of the node, `False` otherwise.
        Raises a ValueError if the argument is used, but it's in a wrong format.

        :param name: a fully qualified name, a local name or a wildcard. The accepted \
        wildcard formats are '*', '*:*', '*:local-name' and '{namespace}*'.
        :param default_namespace: the default namespace for matching element names. \
        The default is no-namespace.
        """
        return False

    def get_child_position(self, child: ta.ChildNodeType) -> int:
        pos = 0
        if self.children:
            for c in self.children:
                if isinstance(child, ElementNode):
                    if c.name == child.name:
                        pos += 1
                elif isinstance(c, child.__class__):
                    pos += 1
                if c is child:
                    break
        return pos


###
# NAMESPACE NODES

class NamespaceNode(XPathNode):
    """
    A class for processing XPath namespace nodes.

    :param prefix: the namespace prefix.
    :param uri: the namespace URI.
    :param parent: the parent element node.
    :param position: the position of the node in the document.
    """
    value: str
    parent: 'ElementNode | None'

    __slots__ = ()

    def __init__(self,
                 prefix: str | None, uri: str,
                 parent: 'ElementNode | None' = None,
                 position: int = 1) -> None:
        self.name = prefix
        self.value = uri
        self.parent = parent
        self.position = position

    def __repr__(self) -> str:
        return '%s(prefix=%r, uri=%r)' % (self.__class__.__name__, self.name, self.value)

    @property
    def prefix(self) -> str | None:
        return self.name

    @property
    def uri(self) -> str:
        return self.value

    @property
    def name_path(self) -> str:
        return self.prefix or _EMPTY_NAME_PATH

    @property
    def path(self) -> str:
        if self.parent is None:
            return '/namespace::{name_path}'
        elif isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/namespace::{self.name_path}"
        return f"/namespace::{self.name_path}"

    @property
    def node_kind(self) -> str:
        return 'namespace'

    @property
    def node_name(self) -> QName | None:
        return None if not self.name else QName(None, self.name)

    @property
    def string_value(self) -> str:
        return self.value

    @property
    def iter_typed_values(self) -> Iterator[str]:
        yield self.value


###
# ATTRIBUTE NODES

class AttributeNode(XPathNode):
    """
    Base class for XPath attribute nodes, used only for type checking.
    """
    name: str | None
    parent: Optional['ElementNode']
    xsd_type: XsdTypeProtocol | None

    __slots__ = ('xsd_type',)

    def __new__(cls, *args: Any, **kwargs: Any) -> 'AttributeNode':
        if cls is AttributeNode:
            return object.__new__(TextAttributeNode)
        return object.__new__(cls)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r}, value={self.value!r})'

    @property
    def uri_qualified_name(self) -> str | None:
        """The URI qualified name of the attribute."""
        if not self.name:
            return self.name
        elif self.name[0] == '{':
            return f'Q{self.name}'
        else:
            return self.name

    @property
    def name_path(self) -> str:
        return self.uri_qualified_name or _EMPTY_NAME_PATH

    @property
    def path(self) -> str:
        if self.parent is None:
            return f'/@{self.name_path}'
        elif isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/@{self.name_path}"
        return f"/@{self.name_path}"

    @property
    def is_typed(self) -> bool:
        return self.xsd_type is not None

    @property
    def is_list(self) -> bool:
        return self.xsd_type is not None and self.xsd_type.is_list()

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        return self.name == name or '*' in name and match_wildcard(self.name, name)

    @property
    def base_uri(self) -> str | None:
        return self.parent.base_uri if self.parent is not None else None

    @property
    def is_id(self) -> bool:
        return self.name == XML_ID or self.xsd_type is not None and self.xsd_type.is_key()

    @property
    def is_idrefs(self) -> bool:
        if self.xsd_type is None:
            return False
        root_type = self.xsd_type.root_type
        return root_type.name == XSD_IDREF or root_type.name == XSD_IDREFS

    @property
    def node_kind(self) -> str:
        return 'attribute'

    @property
    def string_value(self) -> str:
        raise NotImplementedError()

    @property
    def type_name(self) -> str | None:
        return XSD_UNTYPED_ATOMIC if self.xsd_type is None else self.xsd_type.name

    @property
    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
        value = [x for x in self.iter_typed_values]
        return value[0] if len(value) == 1 else value

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        raise NotImplementedError()


class TextAttributeNode(AttributeNode):
    """
    Class for processing XPath attribute nodes.

    :param name: the attribute name.
    :param value: the string value of the attribute.
    :param parent: the parent element node.
    :param position: the position of the node in the document.
    """
    name: str
    value: str
    parent: 'EtreeElementNode | None'

    __slots__ = ()

    def __init__(self,
                 name: str,
                 value: str,
                 parent: 'EtreeElementNode | None' = None,
                 position: int = 1) -> None:

        self.name = name
        self.value = value
        self.parent = parent
        self.position = position
        self.xsd_type = None

    def __repr__(self) -> str:
        return '%s(name=%r, value=%r)' % (self.__class__.__name__, self.name, self.value)

    @property
    def string_value(self) -> str:
        return self.value

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        if self.parent is not None:
            yield from get_atomic_sequence(self.xsd_type, self.value, self.parent.nsmap)
        else:
            yield from get_atomic_sequence(self.xsd_type, self.value)

    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
        if self.parent is not None:
            self.parent.apply_schema(schema)
        elif (xsd_attribute := schema.get_attribute(self.name)) is not None:
            self.xsd_type = xsd_attribute.type
        else:
            self.xsd_type = None


class SchemaAttributeNode(AttributeNode):
    """A class for processing XML Schema attribute nodes."""
    name: str | None
    value: XsdAttributeProtocol
    parent: 'ElementNode | None'

    __slots__ = ()

    def __init__(self,
                 attr: XsdAttributeProtocol,
                 parent: 'ElementNode | None' = None,
                 position: int = 1) -> None:

        self.name = attr.name
        self.value = attr
        self.parent = parent
        self.position = position
        self.xsd_type = attr.type

    def __repr__(self) -> str:
        return '%s(attr=%r)' % (self.__class__.__name__, self.value)

    @property
    def string_value(self) -> str:
        return str(get_atomic_sequence(self.xsd_type))

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        yield from get_atomic_sequence(self.xsd_type)

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        if not self.name:
            return self.value.is_matching(name, default_namespace)
        elif '*' in name:
            return match_wildcard(self.name, name)
        else:
            return self.name == name

    @property
    def is_schema_node(self) -> bool:
        return True


###
# TEXT NODES

class TextNode(XPathNode):
    """
    A class for processing XPath text nodes. An Element's property
    (elem.text or elem.tail) with a `None` value is not a text node.

    :param content: a string value.
    :param parent: the parent element node.
    :param position: the position of the node in the document.
    """
    name: None
    value: str
    parent: 'ElementNode | None'
    children: None = None

    __slots__ = ()

    def __init__(self,
                 content: str,
                 parent: 'ElementNode | None' = None,
                 position: int = 1) -> None:
        self.name = None
        self.value = content
        self.parent = parent
        self.position = position

        if parent is not None:
            parent.children.append(self)

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, self.value)

    @property
    def content(self) -> str:
        return self.value

    @property
    def path(self) -> str:
        if self.parent is None:
            return '/text()[1]'

        pos = self.parent.get_child_position(self)
        if isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/text()[{pos}]"
        return f"/text()[{pos}]"

    ###
    # Text node accessors

    @property
    def base_uri(self) -> str | None:
        return self.parent.base_uri if self.parent is not None else None

    @property
    def node_kind(self) -> str:
        return 'text'

    @property
    def string_value(self) -> str:
        return self.value

    @property
    def type_name(self) -> str | None:
        return XSD_UNTYPED_ATOMIC

    @property
    def typed_value(self) -> UntypedAtomic:
        return UntypedAtomic(self.value)

    @property
    def iter_typed_values(self) -> Iterator[UntypedAtomic]:
        yield UntypedAtomic(self.value)


###
# COMMENT NODES

class CommentNode(XPathNode):
    """
    A class for processing XPath comment nodes.

    :param content: the wrapped Comment Element or a string.
    :param parent: the parent node.
    :param position: the position of the node in the document.
    """
    name: None
    value: CommentType

    __slots__ = ()

    def __init__(self,
                 content: CommentType | str,
                 parent: ta.ParentNodeType | None = None,
                 position: int = 1) -> None:

        self.name = None
        if isinstance(content, str):
            self.value = ElementTree.Comment(content)
        else:
            self.value = content

        self.parent = parent
        self.position = position

        if parent is not None:
            parent.children.append(self)
            parent.tree.elements[self.value] = self

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, self.value.text or '')

    @property
    def content(self) -> CommentType:
        return self.value

    elem = content

    @property
    def path(self) -> str:
        if self.parent is None:
            return '/comment()[1]'

        pos = self.parent.get_child_position(self)
        if isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/comment()[{pos}]"
        return f"/comment()[{pos}]"

    @property
    def base_uri(self) -> str | None:
        return self.parent.base_uri if self.parent is not None else None

    @property
    def node_kind(self) -> str:
        return 'comment'

    @property
    def string_value(self) -> str:
        return self.value.text or ''

    @property
    def typed_value(self) -> str:
        return self.string_value

    @property
    def iter_typed_values(self) -> Iterator[str]:
        yield self.string_value


###
# PROCESSING INSTRUCTION NODES

class ProcessingInstructionNode(XPathNode):
    """
    A class for XPath processing instructions nodes.

    :param target: the wrapped Processing Instruction object or a string.
    :param content: an optional string, used if *target* is a string.
    :param parent: the parent element node.
    :param position: the position of the node in the document.
    """
    name: str
    value: ProcessingInstructionType

    __slots__ = ()

    def __init__(self,
                 target: str | ProcessingInstructionType,
                 content: str | None = None,
                 parent: ta.ParentNodeType | None = None,
                 position: int = 1) -> None:

        if isinstance(target, str):
            self.name = target
            self.value = ElementTree.ProcessingInstruction(self.name, content)
        else:
            if hasattr(target, 'target'):
                self.name = cast(str, target.target)  # lxml PI
            else:
                self.name = (target.text or '').partition(' ')[0]
            self.value = target

        self.parent = parent
        self.position = position

        if parent is not None:
            parent.children.append(self)
            parent.tree.elements[self.value] = self

    def __repr__(self) -> str:
        return '%s(target=%r, content=%r)' % (self.__class__.__name__, self.name, self.content)

    @property
    def target(self) -> str:
        return self.name

    @property
    def content(self) -> str:
        if hasattr(self.value, 'target'):
            return self.value.text or ''
        else:
            return (self.value.text or '').partition(' ')[-1]

    @property
    def elem(self) -> ProcessingInstructionType:
        return self.value

    @property
    def path(self) -> str:
        if self.parent is None:
            return '/processing-instruction({self.name})[1]'

        pos = self.parent.get_child_position(self)
        if isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/processing-instruction({self.name})[{pos}]"
        return f"/processing-instruction({self.name})[{pos}]"

    @property
    def base_uri(self) -> str | None:
        return self.parent.base_uri if self.parent is not None else None

    @property
    def node_kind(self) -> str:
        return 'processing-instruction'

    @property
    def node_name(self) -> QName:
        return QName(None, self.name)

    @property
    def string_value(self) -> str:
        return self.content

    @property
    def typed_value(self) -> str:
        return self.content

    @property
    def iter_typed_values(self) -> Iterator[str]:
        yield self.content

    text = string_value


###
# ELEMENT NODES

class ElementNode(XPathNode):
    """
    Base class for XPath element nodes, used only for type checking. Element nodes
    use lazy properties to diminish the average load for a tree processing.
    """
    name: str | None
    value: object
    tree: XPathNodeTree
    children: list[ta.ChildNodeType]
    parent: ta.ParentNodeType | None
    xsd_type: XsdTypeProtocol | None
    xsd_element: XsdElementProtocol | None
    _nsmap: ta.NsmapType | ta.NamespacesType | None

    # Lazy protected attributes
    _namespace_nodes: list[NamespaceNode]
    _attributes: list[AttributeNode]

    __slots__ = ('children', 'tree', 'xsd_type', 'xsd_element',
                 '_nsmap', '_namespace_nodes', '_attributes')

    def __new__(cls, *args: Any, **kwargs: Any) -> 'ElementNode':
        if cls is ElementNode:
            return object.__new__(EtreeElementNode)
        return object.__new__(cls)

    def __repr__(self) -> str:
        return '%s(elem=%r)' % (self.__class__.__name__, self.value)

    def __getitem__(self, i: int | slice) -> ta.ChildNodeType | list[ta.ChildNodeType]:
        return self.children[i]

    def __len__(self) -> int:
        return len(self.children)

    def __iter__(self) -> Iterator[ta.ChildNodeType]:
        yield from self.children

    @property
    def nsmap(self) -> ta.NsmapType | ta.NamespacesType:
        if hasattr(self.value, 'nsmap'):
            return cast(ta.NsmapType, self.value.nsmap)
        elif self._nsmap is not None:
            return self._nsmap
        else:
            return self.tree.namespaces

    @nsmap.setter
    def nsmap(self, nsmap: ta.NsmapType | ta.NamespacesType) -> None:
        self._nsmap = nsmap

    @property
    def uri_qualified_name(self) -> str | None:
        """The URI qualified name of the element."""
        if not self.name:
            return self.name
        elif self.name[0] == '{':
            return f'Q{self.name}'
        else:
            return f'Q{{}}{self.name}'

    @property
    def attributes(self) -> list[AttributeNode]:
        return []

    @property
    def base_uri(self) -> str | None:
        base_uri: str | None = self._uri.strip() if hasattr(self, '_uri') else None
        if self.parent is None:
            return base_uri
        elif base_uri is None:
            return self.parent.base_uri
        else:
            return urljoin(self.parent.base_uri or '', base_uri)

    @property
    def is_id(self) -> bool:
        return self.name == XML_ID or self.xsd_type is not None and self.xsd_type.is_key()

    @property
    def is_idrefs(self) -> bool:
        if self.xsd_type is None:
            return False
        root_type = self.xsd_type.root_type
        return root_type.name == XSD_IDREF or root_type.name == XSD_IDREFS

    @property
    def namespace_nodes(self) -> list[NamespaceNode]:
        if not hasattr(self, '_namespace_nodes'):
            # Lazy generation of namespace nodes of the element
            position = self.position + 1
            self._namespace_nodes = [NamespaceNode('xml', XML_NAMESPACE, self, position)]
            position += 1
            if self.nsmap:
                for pfx, uri in self.nsmap.items():
                    if pfx != 'xml':
                        self._namespace_nodes.append(NamespaceNode(pfx, uri, self, position))
                        position += 1

        return self._namespace_nodes

    @property
    def nilled(self) -> bool:
        return False

    @property
    def node_kind(self) -> str:
        return 'element'

    @property
    def string_value(self) -> str:
        raise NotImplementedError()

    @property
    def type_name(self) -> str | None:
        return XSD_UNTYPED if self.xsd_type is None else self.xsd_type.name

    @property
    def typed_value(self) -> list[ta.AtomicType] | ta.AtomicType:
        value = [x for x in self.iter_typed_values]
        return value[0] if len(value) == 1 else value

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        raise NotImplementedError()

    @property
    def is_list(self) -> bool:
        return self.xsd_type is not None and self.xsd_type.is_list()

    @property
    def uri(self) -> str | None:
        return self.tree.uri

    @uri.setter
    def uri(self, uri: str) -> None:
        self.tree.uri = uri

    @property
    def schema(self) -> Optional[ta.SchemaProxyType]:
        return self.tree.schema

    @schema.setter
    def schema(self, schema: ta.SchemaProxyType) -> None:
        self.tree.schema = schema

    @property
    def elements(self) -> Optional[ta.ElementMapType]:
        return self.tree.elements

    @elements.setter
    def elements(self, elements: ta.ElementMapType) -> None:
        self.tree.elements = elements

    @property
    def name_path(self) -> str:
        return self.uri_qualified_name or _EMPTY_NAME_PATH

    @property
    def path(self) -> str:
        if self.parent is None:
            return f'/{self.name_path}[1]'

        pos = self.parent.get_child_position(self)
        if isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/{self.name_path}[{pos}]"
        return f"/{self.name_path}[{pos}]"

    @property
    def default_namespace(self) -> str | None:
        if None in self.nsmap:
            return self.nsmap[None]  # type: ignore
        else:
            return self.nsmap.get('')

    @property
    def is_typed(self) -> bool:
        return self.xsd_type is not None

    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
        return

    def clear_types(self) -> None:
        return

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        if self.name is None:
            return False
        elif '*' in name:
            return match_wildcard(self.name, name)
        elif not name:
            return not self.name
        elif name[0] == '{' or not default_namespace:
            return self.name == name
        else:
            return self.name == f'{{{default_namespace}}}{name}'

    def get_element_node(self, elem: ElementType | SchemaElemType) -> ta.TaggedNodeType | None:
        return self.tree.elements.get(elem)

    def get_document_node(self, replace: bool = False, as_parent: bool = True) -> 'DocumentNode':
        """
        Returns a `DocumentNode` for the element node. If the element belongs to a tree that
        already has a document root, returns the document, otherwise creates a dummy document.

        :param replace: if `True` the root element of the tree is replaced by the \
        document node. This is usually useful for extended data models (more element \
        children, text nodes). Default is `False`.
        :param as_parent: if `True` the root node/s of parent attribute is set with \
        the dummy document node, otherwise is set to `None`.
        """
        raise NotImplementedError()

    def iter(self) -> Iterator[XPathNode]:
        """Iterates the tree building lazy components."""
        yield self
        yield from self.namespace_nodes
        yield from self.attributes

        for child in self:
            if isinstance(child, ElementNode):
                yield from child.iter()
            else:
                yield child

    iter_document = iter  # For backward compatibility

    def iter_lazy(self) -> Iterator[XPathNode]:
        """Iterates the tree not including the not built lazy components."""
        yield self

        iterators: deque[Any] = deque()  # slightly faster than list()
        children: Iterator[Any] = iter(self.children)

        if hasattr(self, '_namespace_nodes'):
            yield from self._namespace_nodes
        if hasattr(self, '_attributes'):
            yield from self._attributes

        while True:
            for child in children:
                yield child

                if isinstance(child, ElementNode):
                    if hasattr(child, '_namespace_nodes'):
                        yield from child._namespace_nodes
                    if hasattr(child, '_attributes'):
                        yield from child._attributes

                    if child.children:
                        iterators.append(children)
                        children = iter(child.children)
                        break
            else:
                try:
                    children = iterators.pop()
                except IndexError:
                    return

    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
        if with_self:
            yield self

        iterators: deque[Any] = deque()
        children: Iterator[Any] = iter(self.children)

        while True:
            for child in children:
                yield child

                if isinstance(child, ElementNode) and child.children:
                    iterators.append(children)
                    children = iter(child.children)
                    break
            else:
                try:
                    children = iterators.pop()
                except IndexError:
                    return


class EtreeElementNode(ElementNode):
    """
    XPath element nodes for wrapping ElementTree elements.

    :param elem: the wrapped Element.
    :param parent: the parent document node or element node.
    :param position: the position of the node in the document.
    """
    name: str
    value: ElementType

    __slots__ = ('xsd_type', 'xsd_element')

    def __init__(self,
                 elem: ElementType,
                 parent: ta.ParentNodeType | None = None,
                 position: int = 1,
                 nsmap: ta.NsmapType | ta.NamespacesType | None = None) -> None:

        assert not callable(elem.tag)
        self.name = elem.tag
        self.value = elem
        self.parent = parent
        self.position = position
        self.children = []
        self.xsd_type = self.xsd_element = None
        self._nsmap = nsmap

        if parent is not None:
            self.tree = parent.tree
            parent.children.append(self)
        else:
            self.tree = XPathNodeTree(self)

        self.tree.elements[elem] = self

    @property
    def content(self) -> ElementType:
        return self.value

    elem = content

    @property
    def attributes(self) -> list[AttributeNode]:
        if not hasattr(self, '_attributes'):
            nsmap = self.nsmap
            position = self.position + len(nsmap) + int('xml' not in nsmap) + 1
            self._attributes = [
                TextAttributeNode(name, value, self, pos)
                for pos, (name, value) in enumerate(self.value.attrib.items(), position)
            ]

            if self.xsd_type is None or (schema := self.tree.schema) is None:
                return self._attributes
            elif not schema.is_fully_valid():
                any_simple_type = schema.get_type(XSD_ANY_SIMPLE_TYPE)
                for attr in self._attributes:
                    attr.xsd_type = any_simple_type
                return self._attributes
            elif self.xsd_element is None or XSI_TYPE in self.value.attrib:
                xsd_type = self.xsd_type
            else:
                xsd_type = cast(XsdTypeProtocol, self.xsd_element.type)

            if hasattr(xsd_type, 'attributes'):
                attributes = xsd_type.attributes
                for attr in self._attributes:
                    assert attr.name is not None
                    if attr.name.startswith('{http://www.w3.org/2001/XMLSchema-instance}'):
                        attr.xsd_type = schema.get_type(XSD_ANY_ATOMIC_TYPE)
                    if attr.name in attributes:
                        attr.xsd_type = attributes[attr.name].type
                    elif None in attributes and attributes[None].is_matching(attr.name):
                        xsd_attribute = schema.get_attribute(attr.name)
                        if xsd_attribute is not None:
                            attr.xsd_type = xsd_attribute.type

                # Add missing attributes with a default value, at the same
                # position of the last attribute.
                position = position + len(self.value.attrib)
                for name, xsd_attribute in attributes.items():
                    if name is not None and name not in self.value.attrib:
                        value = getattr(xsd_attribute, 'value_constraint', None)
                        if value is not None:
                            attr = TextAttributeNode(name, value, self, position)
                            attr.xsd_type = xsd_attribute.type
                            self._attributes.append(attr)

        return self._attributes

    @property
    def base_uri(self) -> str | None:
        base_uri = self.value.get(XML_BASE)
        if isinstance(base_uri, str):
            base_uri = base_uri.strip()
        elif base_uri is not None:
            base_uri = ''
        elif hasattr(self, '_uri'):
            base_uri = self._uri.strip()

        if self.parent is None:
            return base_uri
        elif base_uri is None:
            return self.parent.base_uri
        else:
            return urljoin(self.parent.base_uri or '', base_uri)

    @property
    def nilled(self) -> bool:
        return self.value.get(XSI_NIL) in ('true', '1')

    @property
    def string_value(self) -> str:
        if self.xsd_type is not None and self.xsd_type.is_element_only():
            # Element-only text content is normalized
            return ''.join(etree_iter_strings(self.value, normalize=True))

        return ''.join(etree_iter_strings(self.value)) or \
            getattr(self.xsd_element, 'value_constraint', None) or ''

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        if self.xsd_type is None or \
                self.xsd_type.name in _XSD_SPECIAL_TYPES or \
                self.xsd_type.has_mixed_content():
            yield UntypedAtomic(''.join(etree_iter_strings(self.value)))
        elif self.xsd_type.is_element_only():
            return
        elif self.value.get(XSI_NIL) and getattr(self.xsd_type.parent, 'nillable', None):
            return
        elif self.value.text is not None:
            yield from get_atomic_sequence(self.xsd_type, self.value.text, self.nsmap)
        elif self.value.get(XSI_NIL) in ('1', 'true'):
            yield ''
        else:
            value = getattr(self.xsd_element, 'value_constraint', None)
            yield from get_atomic_sequence(self.xsd_type, value or '')

    @property
    def compat_string_value(self) -> str:
        return ''.join(etree_iter_strings(self.value))

    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
        if self.tree.schema is schema and not schema.is_assertion_based():
            return
        self.tree.schema = schema

        if not schema.is_fully_valid():
            element_type = schema.get_type(XSD_ANY_TYPE)
            for node in self.iter_descendants(with_self=True):
                if isinstance(node, EtreeElementNode):
                    node.xsd_type = element_type
                    if hasattr(node, '_attributes'):
                        delattr(node, '_attributes')
            return

        if schema.base_element is not None:
            self.xsd_element = schema.base_element
            xsd_types: list[XsdTypeProtocol | None] = [schema.base_element.type]
            children: Iterator[Any] = iter(self)
            if schema.is_assertion_based():
                self.xsd_type = schema.get_type(XSD_ANY_TYPE)
            else:
                self.xsd_type = schema.base_element.type

            if hasattr(self, '_attributes'):
                delattr(self, '_attributes')
        else:
            root_node: ta.ParentNodeType = self
            while isinstance(root_node.parent, EtreeElementNode):
                root_node = root_node.parent

            xsd_types = [None]
            children = iter((root_node,))

            if hasattr(root_node, '_attributes'):
                delattr(root_node, '_attributes')

        iterators: deque[Any] = deque()
        element_match_cache: defaultdict[
            int, dict[str | None, XsdElementProtocol | None]
        ] = defaultdict(dict)
        while True:
            for node in children:
                if not isinstance(node, EtreeElementNode):
                    continue
                elif XSI_TYPE in node.value.attrib:
                    xsi_type = cast(str, node.value.attrib[XSI_TYPE])
                    try:
                        type_name = get_expanded_name(xsi_type, node.nsmap)
                    except (KeyError, TypeError):
                        node.clear_types()
                        continue
                    else:
                        xsd_type = schema.get_type(type_name)
                else:
                    if xsd_types[-1] is None:
                        xsd_element = schema.get_element(node.name)
                    elif (content := xsd_types[-1].model_group) is None:
                        xsd_element = None
                    elif node.name in (sub_cache := element_match_cache[id(content)]):
                        xsd_element = sub_cache[node.name]
                    else:
                        for xsd_element in content.iter_elements():
                            if xsd_element.is_matching(node.name):
                                if xsd_element.name != node.name:
                                    # a wildcard or a substitute
                                    xsd_element = schema.get_element(node.name)
                                sub_cache[node.name] = xsd_element
                                break
                        else:
                            xsd_element = None

                    xsd_type = getattr(xsd_element, 'type', None)
                    node.xsd_element = xsd_element

                if xsd_type is None:
                    node.clear_types()
                    continue

                node.xsd_type = xsd_type
                if hasattr(node, '_attributes'):
                    delattr(node, '_attributes')

                if len(node.value):
                    xsd_types.append(xsd_type)
                    iterators.append(children)
                    children = iter(node)
                    break
            else:
                try:
                    children = iterators.pop()
                    xsd_types.pop()
                except IndexError:
                    return

    def clear_types(self) -> None:
        """Clear XSD types for element node subtree."""
        for node in self.iter_descendants(with_self=True):
            if isinstance(node, EtreeElementNode):
                node.xsd_type = None
                node.xsd_element = None
                if hasattr(node, '_attributes'):
                    delattr(node, '_attributes')

    @property
    def is_typed(self) -> bool:
        return self.xsd_type is not None

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        if '*' in name:
            return match_wildcard(self.name, name)
        elif not name:
            return not self.name
        elif hasattr(self.value, 'type') and hasattr(self.value, 'is_matching'):
            return cast(XsdElementProtocol, self.value).is_matching(name, default_namespace)
        elif name[0] == '{' or not default_namespace:
            return self.name == name
        else:
            return self.name == f'{{{default_namespace}}}{name}'

    def get_document_node(self, replace: bool = False, as_parent: bool = True) -> 'DocumentNode':
        """
        Returns a `DocumentNode` for the element node. If the element belongs to a tree that
        already has a document root, returns the document, otherwise creates a dummy document
        if the element node wraps an Element of an ElementTree structure or return `None`.

        :param replace: if `True` the root element of the tree is replaced by the \
        document node. This is usually useful for extended data models (more element \
        children, text nodes). Default is `False`.
        :param as_parent: if `True` the root node/s of parent attribute is set with \
        the dummy document node, otherwise is set to `None`.
        """
        if isinstance(self.tree.root_node, DocumentNode):
            return self.tree.root_node

        root_node = self.tree.root_node
        assert isinstance(root_node, EtreeElementNode)

        if root_node.value.__class__.__module__ not in ('lxml.etree', 'lxml.html'):
            etree = ElementTree
        else:
            etree = importlib.import_module('lxml.etree')

        document_node = object.__new__(EtreeDocumentNode)
        document_node.parent = None
        document_node.tree = root_node.tree
        if as_parent:
            document_node.tree.root_node = document_node

        if replace:
            document = etree.ElementTree()
            if sum(isinstance(x, ElementNode) for x in root_node.children) == 1:
                for child in root_node.children:
                    if isinstance(child, ElementNode):
                        document = etree.ElementTree(cast(ElementTree.Element, child.value))
                        break

            document_node.value = document
            document_node.position = root_node.position
            document_node.children = root_node.children

            for child in document_node.children:
                child.parent = document_node if as_parent else None

            root_node.tree.elements.pop(root_node.value)
            del root_node

        else:
            document_node.value = etree.ElementTree(cast(ElementTree.Element, root_node.value))
            document_node.position = root_node.position - 1
            document_node.children = [root_node]

            if as_parent:
                root_node.parent = document_node

        return document_node


###
# Specialized element nodes

class LazyElementNode(EtreeElementNode):
    """
    A fully lazy element node, slower but better if the node has not
    to be used in a document context. The node extends descendants but
    does not record positions and a map of elements.
    """
    __slots__ = ()

    def __iter__(self) -> Iterator[ta.ChildNodeType]:
        if not self.children:
            if self.value.text is not None:
                TextNode(self.value.text, self)
            if len(self.value):
                for elem in self.value:
                    if not callable(elem.tag):
                        LazyElementNode(elem, self)
                    elif elem.tag.__name__ == 'Comment':
                        CommentNode(elem, self)
                    else:
                        ProcessingInstructionNode(elem, parent=self)

                    if elem.tail is not None:
                        TextNode(elem.tail, self)

        yield from self.children

    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
        if with_self:
            yield self

        for child in self:
            if isinstance(child, ElementNode):
                yield from child.iter_descendants()
            else:
                yield child


class SchemaElementNode(ElementNode):
    """
    An element node class for wrapping the XSD schema and its elements.
    The resulting structure can be a tree or a set of disjoint trees.
    With more roots only one of them is the schema node.
    """
    value: SchemaElemType
    ref: 'SchemaElementNode | None'

    __slots__ = ('ref',)

    def __init__(self,
                 elem: SchemaElemType,
                 parent: ta.ParentNodeType | None = None,
                 position: int = 1,
                 nsmap: ta.NsmapType | ta.NamespacesType | None = None):

        self.name = elem.name
        self.value = elem
        self.parent = parent
        self.position = position
        self.children = []
        self.ref = None

        if hasattr(elem, 'type'):
            self.xsd_element = cast(XsdElementProtocol, elem)
            self.xsd_type = self.xsd_element.type
        else:
            self.xsd_element = self.xsd_type = None

        self._nsmap = nsmap

        try:
            self.tree = parent.tree  # type: ignore[union-attr, unused-ignore]
        except AttributeError:
            self.tree = XPathNodeTree(self)
        else:
            parent.children.append(self)  # type: ignore[union-attr, unused-ignore]

        self.tree.elements[elem] = self

    def __iter__(self) -> Iterator[ta.ChildNodeType]:
        if self.ref is None:
            yield from self.children
        else:
            yield from self.ref.children

    @property
    def content(self) -> SchemaElemType:
        return self.value

    elem = content

    @property
    def path(self) -> str:
        if not hasattr(self, 'type'):
            return '/'
        return super().path

    @property
    def is_schema_node(self) -> bool:
        return True

    def match_name(self, name: str, default_namespace: str | None = None) -> bool:
        if '*' in name:
            return match_wildcard(self.name, name)
        elif not name:
            return not self.name
        elif hasattr(self.value, 'type'):
            return self.value.is_matching(name, default_namespace)
        else:
            return self.value.tag == name  # a schema

    @property
    def attributes(self) -> list[AttributeNode]:
        if not hasattr(self, '_attributes'):
            nsmap = self.tree.namespaces
            position = self.position + len(nsmap) + int('xml' not in nsmap) + 1
            self._attributes = [
                SchemaAttributeNode(attr, self, pos)
                for pos, (_, attr) in enumerate(self.value.attrib.items(), position)
            ]
        return self._attributes

    @property
    def base_uri(self) -> str | None:
        base_uri: str | None = self._uri.strip() if hasattr(self, '_uri') else None
        if self.parent is None:
            return base_uri
        elif base_uri is None:
            return self.parent.base_uri
        else:
            return urljoin(self.parent.base_uri or '', base_uri)

    @property
    def type_name(self) -> str | None:
        if (xsd_type := getattr(self.value, 'type', None)) is not None:
            return cast(str | None, xsd_type.name)
        return None

    @property
    def string_value(self) -> str:
        if not hasattr(self.value, 'type'):
            return ''
        for item in get_atomic_sequence(self.xsd_type):
            return str(item)
        return ''

    @property
    def iter_typed_values(self) -> Iterator[ta.AtomicType]:
        yield from get_atomic_sequence(self.xsd_type)

    def iter(self) -> Iterator[XPathNode]:
        yield self

        iterators: list[Any] = []
        children: Iterator[Any] = iter(self.children)

        if hasattr(self, '_namespace_nodes'):
            yield from self._namespace_nodes
        if hasattr(self, '_attributes'):
            yield from self._attributes

        elements = {self}
        while True:
            for child in children:
                if child in elements:
                    continue
                yield child
                elements.add(child)

                if isinstance(child, ElementNode):
                    if hasattr(child, '_namespace_nodes'):
                        yield from child._namespace_nodes
                    if hasattr(child, '_attributes'):
                        yield from child._attributes

                    if child.children:
                        iterators.append(children)
                        children = iter(child.children)
                        break
            else:
                try:
                    children = iterators.pop()
                except IndexError:
                    return

    def iter_descendants(self, with_self: bool = True) -> Iterator[ta.ChildNodeType]:
        if with_self:
            yield self

        iterators: list[Any] = []
        children: Iterator[Any] = iter(self.children)

        elements = {self}
        while True:
            for child in children:
                if child.ref is not None:
                    child = child.ref

                if child in elements:
                    continue
                yield child
                elements.add(child)

                if child.children:
                    iterators.append(children)
                    children = iter(child.children)
                    break
            else:
                try:
                    children = iterators.pop()
                except IndexError:
                    return


###
# DOCUMENT NODES

class DocumentNode(XPathNode):
    """
    Base class for all XPath document nodes.
    """
    name: None
    value: object
    parent: None
    children: list[ta.ChildNodeType]
    tree: XPathNodeTree

    __slots__ = ('children', 'tree')

    def __new__(cls, *args: Any, **kwargs: Any) -> 'DocumentNode':
        if cls is DocumentNode:
            return object.__new__(EtreeDocumentNode)
        return object.__new__(cls)

    def __repr__(self) -> str:
        return '%s(document=%r)' % (self.__class__.__name__, self.value)

    def __getitem__(self, i: int | slice) -> ta.ChildNodeType | list[ta.ChildNodeType]:
        return self.children[i]

    def __len__(self) -> int:
        return len(self.children)

    def __iter__(self) -> Iterator[ta.ChildNodeType]:
        yield from self.children

    @property
    def document(self) -> object:
        return self.value

    @property
    def uri(self) -> str | None:
        return self.tree.uri

    @property
    def elements(self) -> dict[object, ta.TaggedNodeType]:
        return self.tree.elements

    @property
    def path(self) -> str:
        return '/'

    def getroot(self) -> ElementNode:
        for child in self.children:
            if isinstance(child, ElementNode):
                return child
        raise ElementPathRuntimeError("Missing document root")

    def get_element_node(self, elem: ElementProtocol) -> ta.TaggedNodeType | None:
        return self.elements.get(elem)

    def iter(self) -> Iterator[XPathNode]:
        yield self

        for e in self.children:
            if isinstance(e, ElementNode):
                yield from e.iter()
            else:
                yield e

    iter_document = iter

    def iter_lazy(self) -> Iterator[XPathNode]:
        yield self

        for e in self.children:
            if isinstance(e, ElementNode):
                yield from e.iter_lazy()
            else:
                yield e

    def iter_descendants(self, with_self: bool = True) \
            -> Iterator['DocumentNode | ta.ChildNodeType']:
        if with_self:
            yield self

        for e in self.children:
            if isinstance(e, ElementNode):
                yield from e.iter_descendants()
            else:
                yield e

    @property
    def is_typed(self) -> bool:
        for child in self.children:
            if isinstance(child, ElementNode):
                return child.is_typed
        else:
            return False

    def apply_schema(self, schema: ta.SchemaProxyType) -> None:
        for child in self.children:
            if isinstance(child, ElementNode):
                child.apply_schema(schema)

    def clear_types(self) -> None:
        for child in self.children:
            if isinstance(child, ElementNode):
                child.clear_types()

    @property
    def is_extended(self) -> bool:
        """
        Returns `True` if the document node can't be represented with an
        ElementTree structure, `False` otherwise.
        """
        if not self.children:
            raise ElementPathRuntimeError("Missing document root")
        return len(self.children) > 1 or not isinstance(self.children[0], ElementNode)

    @property
    def base_uri(self) -> str | None:
        return self.uri.strip() if self.uri is not None else None

    @property
    def document_uri(self) -> str | None:
        if self.uri is not None and is_absolute_uri(self.uri):
            return self.uri.strip()
        else:
            return None

    @property
    def node_kind(self) -> str:
        return 'document'

    @property
    def string_value(self) -> str:
        raise NotImplementedError()

    @property
    def typed_value(self) -> ta.AtomicType:
        return UntypedAtomic(self.string_value)

    @property
    def iter_typed_values(self) -> Iterator[UntypedAtomic]:
        yield UntypedAtomic(self.string_value)


class EtreeDocumentNode(DocumentNode):
    """
    A class for ElementTree document nodes.

    :param document: the wrapped ElementTree instance.
    :param uri: the document URI.
    :param position: the position of the node in the document, usually 1, \
    or 0 for lxml standalone root elements with siblings.
    """
    value: DocumentType

    __slots__ = ()

    def __init__(self, document: DocumentType,
                 uri: str | None = None,
                 position: int = 1) -> None:

        self.value = document
        self.name = None
        self.parent = None
        self.position = position
        self.children = []
        self.tree = XPathNodeTree(self, uri=uri)

    @property
    def document(self) -> DocumentType:
        return self.value

    @property
    def string_value(self) -> str:
        if not self.children:
            # Fallback for not built documents
            root = self.value.getroot()
            if root is None:
                return ''
            return ''.join(etree_iter_strings(root))
        return ''.join(child.string_value for child in self.children)

    @property
    def compat_string_value(self) -> str:
        if not self.children:
            return self.string_value
        return ''.join(child.compat_string_value for child in self.children)

    @property
    def is_extended(self) -> bool:
        """
        Returns `True` if the document node can't be represented with an
        ElementTree structure, `False` otherwise.
        """
        root = self.value.getroot()
        if root is None or not is_etree_element_instance(root):
            return True
        elif not self.children:
            raise ElementPathRuntimeError("Missing document root")
        elif len(self.children) == 1:
            return not isinstance(self.children[0], ElementNode)
        elif not hasattr(root, 'itersiblings'):
            return True  # an extended xml.etree.ElementTree structure
        elif any(isinstance(x, TextNode) for x in root):
            return True
        else:
            return sum(isinstance(x, ElementNode) for x in root) != 1
